"""Provide the SubmissionListingMixin class."""

from __future__ import annotations

from typing import TYPE_CHECKING

from praw.const import API_PATH
from praw.models.base import PRAWBase
from praw.models.listing.generator import ListingGenerator

if TYPE_CHECKING:
    from collections.abc import Iterator

    from typing_extensions import Unpack

    from praw import models
    from praw.models.listing.generator import ListingGeneratorKwargs


class SubmissionListingMixin(PRAWBase):
    """Adds additional methods pertaining to :class:`.Submission` instances."""

    if TYPE_CHECKING:
        id: str

    def duplicates(self, **generator_kwargs: Unpack[ListingGeneratorKwargs]) -> Iterator[models.Submission]:
        """Return a :class:`.ListingGenerator` for the submission's duplicates.

        Additional keyword arguments are passed in the initialization of
        :class:`.ListingGenerator`.

        Example usage:

        .. code-block:: python

            submission = reddit.submission("5or86n")

            for duplicate in submission.duplicates():
                # process each duplicate
                ...

        .. seealso::

            :meth:`.upvote`

        """
        url = API_PATH["duplicates"].format(submission_id=self.id)
        return ListingGenerator(self._reddit, url, **generator_kwargs)



"""Provide the SubredditStream class."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from praw.models.util import stream_generator

if TYPE_CHECKING:
    from collections.abc import Iterator

    from praw import models


class SubredditStream:
    """Provides submission and comment streams."""

    def __init__(self, subreddit: models.Subreddit) -> None:
        """Initialize a :class:`.SubredditStream` instance.

        :param subreddit: The subreddit associated with the streams.

        """
        self.subreddit = subreddit

    def comments(self, **stream_options: Any) -> Iterator[models.Comment]:
        """Yield new comments as they become available.

        Comments are yielded oldest first. Up to 100 historical comments will initially
        be returned.

        Keyword arguments are passed to :func:`.stream_generator`.

        .. note::

            While PRAW tries to catch all new comments, some high-volume streams,
            especially the r/all stream, may drop some comments.

        For example, to retrieve all new comments made to r/test, try:

        .. code-block:: python

            for comment in reddit.subreddit("test").stream.comments():
                print(comment)

        To only retrieve new submissions starting when the stream is created, pass
        ``skip_existing=True``:

        .. code-block:: python

            subreddit = reddit.subreddit("test")
            for comment in subreddit.stream.comments(skip_existing=True):
                print(comment)

        """
        return stream_generator(self.subreddit.comments, **stream_options)

    def submissions(self, **stream_options: Any) -> Iterator[models.Submission]:
        r"""Yield new :class:`.Submission`\ s as they become available.

        Submissions are yielded oldest first. Up to 100 historical submissions will
        initially be returned.

        Keyword arguments are passed to :func:`.stream_generator`.

        .. note::

            While PRAW tries to catch all new submissions, some high-volume streams,
            especially the r/all stream, may drop some submissions.

        For example, to retrieve all new submissions made to all of Reddit, try:

        .. code-block:: python

            for submission in reddit.subreddit("all").stream.submissions():
                print(submission)

        """
        return stream_generator(self.subreddit.new, **stream_options)



"""Provide the PRAW models."""

from praw.models.auth import Auth
from praw.models.front import Front
from praw.models.helpers import AnnouncementHelper, DraftHelper, LiveHelper, MultiredditHelper, SubredditHelper
from praw.models.inbox import Inbox
from praw.models.list.draft import DraftList
from praw.models.list.moderated import ModeratedList
from praw.models.list.redditor import RedditorList
from praw.models.list.trophy import TrophyList
from praw.models.listing.announcement import AnnouncementListing
from praw.models.listing.domain import DomainListing
from praw.models.listing.generator import ListingGenerator
from praw.models.listing.listing import Listing, ModeratorListing, ModmailConversationsListing
from praw.models.media import EmojiMedia, Media, PostMedia, StylesheetAsset, StylesheetImage, WidgetMedia
from praw.models.mod_action import ModAction
from praw.models.mod_note import ModNote
from praw.models.mod_notes import RedditModNotes, RedditorModNotes, SubredditModNotes
from praw.models.preferences import Preferences
from praw.models.reddit.announcement import Announcement
from praw.models.reddit.collections import Collection
from praw.models.reddit.comment import Comment
from praw.models.reddit.draft import Draft
from praw.models.reddit.emoji import Emoji
from praw.models.reddit.inline_media import InlineGif, InlineImage, InlineMedia, InlineVideo
from praw.models.reddit.live import LiveThread, LiveUpdate
from praw.models.reddit.message import Message, SubredditMessage
from praw.models.reddit.modmail import ModmailAction, ModmailConversation, ModmailMessage
from praw.models.reddit.more import MoreComments
from praw.models.reddit.multi import Multireddit
from praw.models.reddit.poll import PollData, PollOption
from praw.models.reddit.redditor import Redditor
from praw.models.reddit.removal_reasons import RemovalReason
from praw.models.reddit.rules import Rule
from praw.models.reddit.submission import Submission
from praw.models.reddit.subreddit import Subreddit
from praw.models.reddit.user_subreddit import UserSubreddit
from praw.models.reddit.widgets import (
    Button,
    ButtonWidget,
    Calendar,
    CalendarConfiguration,
    CommunityList,
    CustomWidget,
    Hover,
    IDCard,
    Image,
    ImageData,
    ImageWidget,
    Menu,
    MenuLink,
    ModeratorsWidget,
    PostFlairWidget,
    RulesWidget,
    Styles,
    Submenu,
    SubredditWidgets,
    SubredditWidgetsModeration,
    TextArea,
    Widget,
    WidgetModeration,
)
from praw.models.reddit.wikipage import WikiPage
from praw.models.redditors import Redditors
from praw.models.stylesheet import Stylesheet
from praw.models.subreddits import Subreddits
from praw.models.trophy import Trophy
from praw.models.user import User

__all__ = [
    "Announcement",
    "AnnouncementHelper",
    "AnnouncementListing",
    "Auth",
    "Button",
    "ButtonWidget",
    "Calendar",
    "CalendarConfiguration",
    "Collection",
    "Comment",
    "CommunityList",
    "CustomWidget",
    "DomainListing",
    "Draft",
    "DraftHelper",
    "DraftList",
    "Emoji",
    "EmojiMedia",
    "Front",
    "Hover",
    "IDCard",
    "Image",
    "ImageData",
    "ImageWidget",
    "Inbox",
    "InlineGif",
    "InlineImage",
    "InlineMedia",
    "InlineVideo",
    "Listing",
    "ListingGenerator",
    "LiveHelper",
    "LiveThread",
    "LiveUpdate",
    "Media",
    "Menu",
    "MenuLink",
    "Message",
    "ModAction",
    "ModNote",
    "ModeratedList",
    "ModeratorListing",
    "ModeratorsWidget",
    "ModmailAction",
    "ModmailConversation",
    "ModmailConversationsListing",
    "ModmailMessage",
    "MoreComments",
    "Multireddit",
    "MultiredditHelper",
    "PollData",
    "PollOption",
    "PostFlairWidget",
    "PostMedia",
    "Preferences",
    "RedditModNotes",
    "Redditor",
    "RedditorList",
    "RedditorModNotes",
    "Redditors",
    "RemovalReason",
    "Rule",
    "RulesWidget",
    "Styles",
    "Stylesheet",
    "StylesheetAsset",
    "StylesheetImage",
    "Submenu",
    "Submission",
    "Subreddit",
    "SubredditHelper",
    "SubredditMessage",
    "SubredditModNotes",
    "SubredditWidgets",
    "SubredditWidgetsModeration",
    "Subreddits",
    "TextArea",
    "Trophy",
    "TrophyList",
    "User",
    "UserSubreddit",
    "Widget",
    "WidgetMedia",
    "WidgetModeration",
    "WikiPage",
]



"""Provide the Removal Reason class."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, SupportsIndex

from praw.const import API_PATH
from praw.exceptions import ClientException
from praw.models.reddit.base import RedditBase
from praw.util import cachedproperty

if TYPE_CHECKING:
    from collections.abc import Iterator

    import praw
    from praw import models


class RemovalReason(RedditBase):
    """An individual Removal Reason object.

    .. include:: ../../typical_attributes.rst

    =========== ==================================
    Attribute   Description
    =========== ==================================
    ``id``      The ID of the removal reason.
    ``message`` The message of the removal reason.
    ``title``   The title of the removal reason.
    =========== ==================================

    """

    STR_FIELD = "id"

    def __eq__(self, other: str | RemovalReason) -> bool:
        """Return whether the other instance equals the current."""
        if isinstance(other, str):
            return other == str(self)
        return isinstance(other, self.__class__) and str(self) == str(other)

    def __hash__(self) -> int:
        """Return the hash of the current instance."""
        return hash(self.__class__.__name__) ^ hash(str(self))

    def __init__(
        self,
        reddit: praw.Reddit,
        subreddit: models.Subreddit,
        id: str | None = None,
        _data: dict[str, Any] | None = None,
    ) -> None:
        """Initialize a :class:`.RemovalReason` instance.

        :param reddit: An instance of :class:`.Reddit`.
        :param subreddit: An instance of :class:`.Subreddit`.
        :param id: The ID of the removal reason.

        """
        if (id, _data).count(None) != 1:
            msg = "Either id or _data needs to be given."
            raise ValueError(msg)

        if id:
            self.id = id
        self.subreddit = subreddit
        super().__init__(reddit, _data=_data)

    def _fetch(self) -> None:
        for removal_reason in self.subreddit.mod.removal_reasons:
            if removal_reason.id == self.id:
                self.__dict__.update(removal_reason.__dict__)
                super()._fetch()
                return
        msg = f"Subreddit {self.subreddit} does not have the removal reason {self.id}"
        raise ClientException(msg)

    def delete(self) -> None:
        """Delete a removal reason from this subreddit.

        To delete ``"141vv5c16py7d"`` from r/test try:

        .. code-block:: python

            reddit.subreddit("test").mod.removal_reasons["141vv5c16py7d"].delete()

        """
        url = API_PATH["removal_reason"].format(id=self.id, subreddit=self.subreddit)
        self._reddit.delete(url)

    def update(self, *, message: str | None = None, title: str | None = None) -> None:
        """Update the removal reason from this subreddit.

        .. note::

            Existing values will be used for any unspecified arguments.

        :param message: The removal reason's new message.
        :param title: The removal reason's new title.

        To update ``"141vv5c16py7d"`` from r/test try:

        .. code-block:: python

            reddit.subreddit("test").mod.removal_reasons["141vv5c16py7d"].update(
                title="New title", message="New message"
            )

        """
        url = API_PATH["removal_reason"].format(id=self.id, subreddit=self.subreddit)
        data = {
            name: getattr(self, name) if value is None else value
            for name, value in {"message": message, "title": title}.items()
        }
        self._reddit.put(url, data=data)


class SubredditRemovalReasons:
    """Provide a set of functions to a :class:`.Subreddit`'s removal reasons."""

    @cachedproperty
    def _removal_reason_list(self) -> list[RemovalReason]:
        """Get a list of Removal Reason objects.

        :returns: A list of instances of :class:`.RemovalReason`.

        """
        response = self._reddit.get(API_PATH["removal_reasons_list"].format(subreddit=self.subreddit))
        return [
            RemovalReason(self._reddit, self.subreddit, _data=response["data"][reason_id])
            for reason_id in response["order"]
        ]

    def __getitem__(self, reason_id: SupportsIndex) -> RemovalReason:
        """Return the Removal Reason with the ID/number/slice ``reason_id``.

        :param reason_id: The ID or index of the removal reason

        .. note::

            Removal reasons fetched using a specific rule name are lazily loaded, so you
            might have to access an attribute to get all the expected attributes.

        This method is to be used to fetch a specific removal reason, like so:

        .. code-block:: python

            reason_id = "141vv5c16py7d"
            reason = reddit.subreddit("test").mod.removal_reasons[reason_id]
            print(reason)

        You can also use indices to get a numbered removal reason. Since Python uses
        0-indexing, the first removal reason is index 0, and so on.

        .. note::

            Both negative indices and slices can be used to interact with the removal
            reasons.

        :raises: :py:class:`IndexError` if a removal reason of a specific number does
            not exist.

        For example, to get the second removal reason of r/test:

        .. code-block:: python

            reason = reddit.subreddit("test").mod.removal_reasons[1]

        To get the last three removal reasons in a subreddit:

        .. code-block:: python

            reasons = reddit.subreddit("test").mod.removal_reasons[-3:]
            for reason in reasons:
                print(reason)

        """
        if not isinstance(reason_id, str):
            return self._removal_reason_list[reason_id]
        return RemovalReason(self._reddit, self.subreddit, reason_id)

    def __init__(self, subreddit: models.Subreddit) -> None:
        """Initialize a :class:`.SubredditRemovalReasons` instance.

        :param subreddit: The subreddit whose removal reasons to work with.

        """
        self.subreddit = subreddit
        self._reddit = subreddit._reddit

    def __iter__(self) -> Iterator[RemovalReason]:
        """Return a list of Removal Reasons for the subreddit.

        This method is used to discover all removal reasons for a subreddit:

        .. code-block:: python

            for removal_reason in reddit.subreddit("test").mod.removal_reasons:
                print(removal_reason)

        """
        return iter(self._removal_reason_list)

    def add(self, *, message: str, title: str) -> RemovalReason:
        """Add a removal reason to this subreddit.

        :param message: The message associated with the removal reason.
        :param title: The title of the removal reason.

        :returns: The :class:`.RemovalReason` added.

        The message will be prepended with ``Hi u/username,`` automatically.

        To add ``"Test"`` to r/test try:

        .. code-block:: python

            reddit.subreddit("test").mod.removal_reasons.add(title="Test", message="Foobar")

        """
        data = {"message": message, "title": title}
        url = API_PATH["removal_reasons_list"].format(subreddit=self.subreddit)
        reason_id = self._reddit.post(url, data=data)
        return RemovalReason(self._reddit, self.subreddit, reason_id)



"""Provide the Front class."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any
from urllib.parse import urljoin

from praw.models.listing.generator import ListingGenerator
from praw.models.listing.mixins import SubredditListingMixin

if TYPE_CHECKING:
    from collections.abc import Iterator

    import praw
    from praw import models


class Front(SubredditListingMixin):
    """Front is a Listing class that represents the front page."""

    def __init__(self, reddit: praw.Reddit) -> None:
        """Initialize a :class:`.Front` instance."""
        super().__init__(reddit, _data=None)
        self._path = "/"

    def best(self, **generator_kwargs: Any) -> Iterator[models.Submission]:
        """Return a :class:`.ListingGenerator` for best items.

        Additional keyword arguments are passed in the initialization of
        :class:`.ListingGenerator`.

        """
        return ListingGenerator(self._reddit, urljoin(self._path, "best"), **generator_kwargs)



"""Provide the ReportableMixin class."""

from __future__ import annotations

from typing import TYPE_CHECKING

from praw.const import API_PATH

if TYPE_CHECKING:
    import praw


class ReportableMixin:
    """Interface for :class:`.RedditBase` classes that can be reported."""

    if TYPE_CHECKING:
        # Provided by the host class (:class:`.RedditBase`).
        _reddit: praw.Reddit

        @property
        def fullname(self) -> str: ...  # noqa: D102

    def report(self, reason: str) -> None:
        """Report this object to the moderators of its subreddit.

        :param reason: The reason for reporting.

        :raises: :class:`.RedditAPIException` if ``reason`` is longer than 100
            characters.

        Example usage:

        .. code-block:: python

            submission = reddit.submission("5or86n")
            submission.report("report reason")

            comment = reddit.comment("dxolpyc")
            comment.report("report reason")

        """
        self._reddit.post(API_PATH["report"], data={"id": self.fullname, "reason": reason})



"""Provide the RedditorListingMixin class."""

from __future__ import annotations

from typing import TYPE_CHECKING
from urllib.parse import urljoin

from praw.models.listing.generator import ListingGenerator
from praw.models.listing.mixins.base import BaseListingMixin
from praw.util.cache import cachedproperty

if TYPE_CHECKING:
    from collections.abc import Iterator

    from typing_extensions import Unpack

    import praw
    from praw import models
    from praw.models.listing.generator import ListingGeneratorKwargs


class RedditorListingMixin(BaseListingMixin):
    """Adds additional methods pertaining to :class:`.Redditor` instances."""

    @cachedproperty
    def comments(self) -> SubListing:
        r"""Provide an instance of :class:`.SubListing` for comment access.

        For example, to output the first line of all new comments by u/spez try:

        .. code-block:: python

            for comment in reddit.redditor("spez").comments.new(limit=None):
                print(comment.body.split("\\n", 1)[0][:79])

        """
        return SubListing(self._reddit, self._path, "comments")

    @cachedproperty
    def overview(self) -> SubListing:
        r"""Provide an instance of :class:`.SubListing` for overview access.

        The overview combines a Redditor's comments and submissions, mirroring the user
        overview page on Reddit.

        .. note::

            This is the same listing produced by calling a sort method directly on the
            :class:`.Redditor` instance, so ``reddit.redditor("spez").overview.new()``
            and ``reddit.redditor("spez").new()`` yield the same items. Use
            :attr:`.comments` or :attr:`.submissions` to restrict the listing to a
            single type.

        For example, to output the first line of all top items by u/spez try:

        .. code-block:: python

            for item in reddit.redditor("spez").overview.top(time_filter="all"):
                print(str(item)[:79])

        """
        return SubListing(self._reddit, self._path, "overview")

    @cachedproperty
    def submissions(self) -> SubListing:
        """Provide an instance of :class:`.SubListing` for submission access.

        For example, to output the title's of top 100 of all time submissions for u/spez
        try:

        .. code-block:: python

            for submission in reddit.redditor("spez").submissions.top(time_filter="all"):
                print(submission.title)

        """
        return SubListing(self._reddit, self._path, "submitted")

    def downvoted(
        self, **generator_kwargs: Unpack[ListingGeneratorKwargs]
    ) -> Iterator[models.Comment | models.Submission]:
        """Return a :class:`.ListingGenerator` for items the user has downvoted.

        :returns: A :class:`.ListingGenerator` object which yields :class:`.Comment` or
            :class:`.Submission` objects the user has downvoted.

        :raises: ``prawcore.Forbidden`` if the user is not authorized to access the
            list.

            .. note::

                Since this function returns a :class:`.ListingGenerator` the exception
                may not occur until sometime after this function has returned.


        Additional keyword arguments are passed in the initialization of
        :class:`.ListingGenerator`.

        For example, to get all downvoted items of the authenticated user:

        .. code-block:: python

            for item in reddit.user.me().downvoted():
                print(item.id)

        """
        return ListingGenerator(self._reddit, urljoin(self._path, "downvoted"), **generator_kwargs)

    def hidden(
        self, **generator_kwargs: Unpack[ListingGeneratorKwargs]
    ) -> Iterator[models.Comment | models.Submission]:
        """Return a :class:`.ListingGenerator` for items the user has hidden.

        :returns: A :class:`.ListingGenerator` object which yields :class:`.Comment` or
            :class:`.Submission` objects the user has hid.

        :raises: ``prawcore.Forbidden`` if the user is not authorized to access the
            list.

            .. note::

                Since this function returns a :class:`.ListingGenerator` the exception
                may not occur until sometime after this function has returned.


        Additional keyword arguments are passed in the initialization of
        :class:`.ListingGenerator`.

        For example, to get all hidden items of the authenticated user:

        .. code-block:: python

            for item in reddit.user.me().hidden():
                print(item.id)

        """
        return ListingGenerator(self._reddit, urljoin(self._path, "hidden"), **generator_kwargs)

    def saved(self, **generator_kwargs: Unpack[ListingGeneratorKwargs]) -> Iterator[models.Comment | models.Submission]:
        """Return a :class:`.ListingGenerator` for items the user has saved.

        :returns: A :class:`.ListingGenerator` object which yields :class:`.Comment` or
            :class:`.Submission` objects the user has saved.

        :raises: ``prawcore.Forbidden`` if the user is not authorized to access the
            list.

            .. note::

                Since this function returns a :class:`.ListingGenerator` the exception
                may not occur until sometime after this function has returned.


        Additional keyword arguments are passed in the initialization of
        :class:`.ListingGenerator`.

        For example, to get all saved items of the authenticated user:

        .. code-block:: python

            for item in reddit.user.me().saved(limit=None):
                print(item.id)

        """
        return ListingGenerator(self._reddit, urljoin(self._path, "saved"), **generator_kwargs)

    def upvoted(
        self, **generator_kwargs: Unpack[ListingGeneratorKwargs]
    ) -> Iterator[models.Comment | models.Submission]:
        """Return a :class:`.ListingGenerator` for items the user has upvoted.

        :returns: A :class:`.ListingGenerator` object which yields :class:`.Comment` or
            :class:`.Submission` objects the user has upvoted.

        :raises: ``prawcore.Forbidden`` if the user is not authorized to access the
            list.

            .. note::

                Since this function returns a :class:`.ListingGenerator` the exception
                may not occur until sometime after this function has returned.


        Additional keyword arguments are passed in the initialization of
        :class:`.ListingGenerator`.

        For example, to get all upvoted items of the authenticated user:

        .. code-block:: python

            for item in reddit.user.me().upvoted():
                print(item.id)

        """
        return ListingGenerator(self._reddit, urljoin(self._path, "upvoted"), **generator_kwargs)


class SubListing(BaseListingMixin):
    """Helper class for generating :class:`.ListingGenerator` objects."""

    def __init__(self, reddit: praw.Reddit, base_path: str, subpath: str) -> None:
        """Initialize a :class:`.SubListing` instance.

        :param reddit: An instance of :class:`.Reddit`.
        :param base_path: The path to the object up to this point.
        :param subpath: The additional path to this sublisting.

        """
        super().__init__(reddit, _data=None)
        self._listing_use_sort = True
        self._reddit = reddit
        self._path = urljoin(base_path, subpath)



"""Provide the RisingListingMixin class."""

from __future__ import annotations

from typing import TYPE_CHECKING
from urllib.parse import urljoin

from praw.models.base import PRAWBase
from praw.models.listing.generator import ListingGenerator

if TYPE_CHECKING:
    from collections.abc import Iterator

    from typing_extensions import Unpack

    from praw import models
    from praw.models.listing.generator import ListingGeneratorKwargs


class RisingListingMixin(PRAWBase):
    """Mixes in the rising methods."""

    if TYPE_CHECKING:
        _path: str

    def rising(self, **generator_kwargs: Unpack[ListingGeneratorKwargs]) -> Iterator[models.Submission]:
        """Return a :class:`.ListingGenerator` for rising submissions.

        Additional keyword arguments are passed in the initialization of
        :class:`.ListingGenerator`.

        For example, to get rising submissions for r/test:

        .. code-block:: python

            for submission in reddit.subreddit("test").rising():
                print(submission.title)

        """
        return ListingGenerator(self._reddit, urljoin(self._path, "rising"), **generator_kwargs)



"""Provide the SavableMixin class."""

from __future__ import annotations

from typing import TYPE_CHECKING

from praw.const import API_PATH

if TYPE_CHECKING:
    import praw


class SavableMixin:
    """Interface for :class:`.RedditBase` classes that can be saved."""

    if TYPE_CHECKING:
        # Provided by the host class (:class:`.RedditBase`).
        _reddit: praw.Reddit

        @property
        def fullname(self) -> str: ...  # noqa: D102

    def save(self, *, category: str | None = None) -> None:
        """Save the object.

        :param category: The category to save to. If the authenticated user does not
            have Reddit Premium this value is ignored by Reddit (default: ``None``).

        Example usage:

        .. code-block:: python

            submission = reddit.submission("5or86n")
            submission.save(category="view later")

            comment = reddit.comment("dxolpyc")
            comment.save()

        .. seealso::

            :meth:`.unsave`

        """
        self._reddit.post(API_PATH["save"], data={"category": category, "id": self.fullname})

    def unsave(self) -> None:
        """Unsave the object.

        Example usage:

        .. code-block:: python

            submission = reddit.submission("5or86n")
            submission.unsave()

            comment = reddit.comment("dxolpyc")
            comment.unsave()

        .. seealso::

            :meth:`.save`

        """
        self._reddit.post(API_PATH["unsave"], data={"id": self.fullname})



"""Provide the SubredditStylesheet class."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from praw.const import API_PATH

if TYPE_CHECKING:
    from praw import models


class SubredditStylesheet:
    """Provides a set of stylesheet functions to a :class:`.Subreddit`.

    For example, to add the css data ``.test{color:blue}`` to the existing stylesheet:

    .. code-block:: python

        subreddit = reddit.subreddit("test")
        stylesheet = subreddit.stylesheet()
        stylesheet.stylesheet += ".test{color:blue}"
        subreddit.stylesheet.update(stylesheet.stylesheet)

    """

    def __call__(self) -> models.Stylesheet:
        """Return the :class:`.Subreddit`'s stylesheet.

        To be used as:

        .. code-block:: python

            stylesheet = reddit.subreddit("test").stylesheet()

        """
        url = API_PATH["about_stylesheet"].format(subreddit=self.subreddit)
        return self.subreddit._reddit.get(url)

    def __init__(self, subreddit: models.Subreddit) -> None:
        """Initialize a :class:`.SubredditStylesheet` instance.

        :param subreddit: The :class:`.Subreddit` associated with the stylesheet.

        An instance of this class is provided as:

        .. code-block:: python

            reddit.subreddit("test").stylesheet

        """
        self.subreddit = subreddit

    def _update_structured_styles(self, style_data: dict[str, str | Any]) -> None:
        url = API_PATH["structured_styles"].format(subreddit=self.subreddit)
        self.subreddit._reddit.patch(url, data=style_data)

    def delete_banner(self) -> None:
        """Remove the current :class:`.Subreddit` (redesign) banner image.

        Succeeds even if there is no banner image.

        For example:

        .. code-block:: python

            reddit.subreddit("test").stylesheet.delete_banner()

        """
        data = {"bannerBackgroundImage": ""}
        self._update_structured_styles(data)

    def delete_banner_additional_image(self) -> None:
        """Remove the current :class:`.Subreddit` (redesign) banner additional image.

        Succeeds even if there is no additional image. Will also delete any configured
        hover image.

        For example:

        .. code-block:: python

            reddit.subreddit("test").stylesheet.delete_banner_additional_image()

        """
        data = {"bannerPositionedImage": "", "secondaryBannerPositionedImage": ""}
        self._update_structured_styles(data)

    def delete_banner_hover_image(self) -> None:
        """Remove the current :class:`.Subreddit` (redesign) banner hover image.

        Succeeds even if there is no hover image.

        For example:

        .. code-block:: python

            reddit.subreddit("test").stylesheet.delete_banner_hover_image()

        """
        data = {"secondaryBannerPositionedImage": ""}
        self._update_structured_styles(data)

    def delete_header(self) -> None:
        """Remove the current :class:`.Subreddit` header image.

        Succeeds even if there is no header image.

        For example:

        .. code-block:: python

            reddit.subreddit("test").stylesheet.delete_header()

        """
        url = API_PATH["delete_sr_header"].format(subreddit=self.subreddit)
        self.subreddit._reddit.post(url)

    def delete_image(self, name: str) -> None:
        """Remove the named image from the :class:`.Subreddit`.

        Succeeds even if the named image does not exist.

        For example:

        .. code-block:: python

            reddit.subreddit("test").stylesheet.delete_image("smile")

        """
        url = API_PATH["delete_sr_image"].format(subreddit=self.subreddit)
        self.subreddit._reddit.post(url, data={"img_name": name})

    def delete_mobile_banner(self) -> None:
        """Remove the current :class:`.Subreddit` (redesign) mobile banner.

        Succeeds even if there is no mobile banner.

        For example:

        .. code-block:: python

            subreddit = reddit.subreddit("test")
            subreddit.stylesheet.delete_banner_hover_image()

        """
        data = {"mobileBannerImage": ""}
        self._update_structured_styles(data)

    def delete_mobile_header(self) -> None:
        """Remove the current :class:`.Subreddit` mobile header.

        Succeeds even if there is no mobile header.

        For example:

        .. code-block:: python

            reddit.subreddit("test").stylesheet.delete_mobile_header()

        """
        url = API_PATH["delete_sr_header"].format(subreddit=self.subreddit)
        self.subreddit._reddit.post(url)

    def delete_mobile_icon(self) -> None:
        """Remove the current :class:`.Subreddit` mobile icon.

        Succeeds even if there is no mobile icon.

        For example:

        .. code-block:: python

            reddit.subreddit("test").stylesheet.delete_mobile_icon()

        """
        url = API_PATH["delete_sr_icon"].format(subreddit=self.subreddit)
        self.subreddit._reddit.post(url)

    def update(self, stylesheet: str, *, reason: str | None = None) -> None:
        """Update the :class:`.Subreddit`'s stylesheet.

        :param stylesheet: The CSS for the new stylesheet.
        :param reason: The reason for updating the stylesheet.

        For example:

        .. code-block:: python

            reddit.subreddit("test").stylesheet.update(
                "p { color: green; }", reason="color text green"
            )

        """
        data = {"op": "save", "reason": reason, "stylesheet_contents": stylesheet}
        url = API_PATH["subreddit_stylesheet"].format(subreddit=self.subreddit)
        self.subreddit._reddit.post(url, data=data)

    def upload(self, media: models.StylesheetImage, /, *, name: str) -> dict[str, str]:
        """Upload an image to the :class:`.Subreddit`.

        :param media: The :class:`.StylesheetImage` to upload.
        :param name: The name to use for the image. If an image already exists with the
            same name, it will be replaced.

        :returns: A dictionary containing a link to the uploaded image under the key
            ``img_src``.

        :raises: ``prawcore.TooLarge`` if the overall request body is too large.
        :raises: :class:`.RedditAPIException` if there are other issues with the
            uploaded image. Unfortunately the exception info might not be very specific,
            so try through the website with the same image to see what the problem
            actually might be.

        For example:

        .. code-block:: python

            from praw.models import StylesheetImage

            reddit.subreddit("test").stylesheet.upload(StylesheetImage("img.png"), name="smile")

        """
        return media._upload(self.subreddit, name=name, upload_type="img")

    def upload_banner(self, media: models.StylesheetAsset, /) -> None:
        """Upload an image for the :class:`.Subreddit`'s (redesign) banner image.

        :param media: The :class:`.StylesheetAsset` to upload.

        :raises: ``prawcore.TooLarge`` if the overall request body is too large.
        :raises: :class:`.RedditAPIException` if there are other issues with the
            uploaded image. Unfortunately the exception info might not be very specific,
            so try through the website with the same image to see what the problem
            actually might be.

        For example:

        .. code-block:: python

            from praw.models import StylesheetAsset

            reddit.subreddit("test").stylesheet.upload_banner(StylesheetAsset("banner.png"))

        """
        image_type = "bannerBackgroundImage"
        image_url = media._upload(self.subreddit, image_type=image_type)
        self._update_structured_styles({image_type: image_url})

    def upload_banner_additional_image(
        self,
        media: models.StylesheetAsset,
        /,
        *,
        align: str | None = None,
    ) -> None:
        """Upload an image for the :class:`.Subreddit`'s (redesign) additional image.

        :param media: The :class:`.StylesheetAsset` to upload.
        :param align: Either ``"left"``, ``"centered"``, or ``"right"``. (default:
            ``"left"``).

        :raises: ``prawcore.TooLarge`` if the overall request body is too large.
        :raises: :class:`.RedditAPIException` if there are other issues with the
            uploaded image. Unfortunately the exception info might not be very specific,
            so try through the website with the same image to see what the problem
            actually might be.

        For example:

        .. code-block:: python

            from praw.models import StylesheetAsset

            subreddit = reddit.subreddit("test")
            subreddit.stylesheet.upload_banner_additional_image(StylesheetAsset("banner.png"))

        """
        alignment = {}
        if align is not None:
            if align not in {"left", "centered", "right"}:
                msg = "'align' argument must be either 'left', 'centered', or 'right'"
                raise ValueError(msg)
            alignment["bannerPositionedImagePosition"] = align

        image_type = "bannerPositionedImage"
        image_url = media._upload(self.subreddit, image_type=image_type)
        style_data = {image_type: image_url}
        if alignment:
            style_data.update(alignment)
        self._update_structured_styles(style_data)

    def upload_banner_hover_image(self, media: models.StylesheetAsset, /) -> None:
        """Upload an image for the :class:`.Subreddit`'s (redesign) additional image.

        :param media: The :class:`.StylesheetAsset` to upload.

        Fails if the :class:`.Subreddit` does not have an additional image defined.

        :raises: ``prawcore.TooLarge`` if the overall request body is too large.
        :raises: :class:`.RedditAPIException` if there are other issues with the
            uploaded image. Unfortunately the exception info might not be very specific,
            so try through the website with the same image to see what the problem
            actually might be.

        For example:

        .. code-block:: python

            from praw.models import StylesheetAsset

            subreddit = reddit.subreddit("test")
            subreddit.stylesheet.upload_banner_hover_image(StylesheetAsset("banner.png"))

        """
        image_type = "secondaryBannerPositionedImage"
        image_url = media._upload(self.subreddit, image_type=image_type)
        self._update_structured_styles({image_type: image_url})

    def upload_header(self, media: models.StylesheetImage, /) -> dict[str, str]:
        """Upload an image to be used as the :class:`.Subreddit`'s header image.

        :param media: The :class:`.StylesheetImage` to upload.

        :returns: A dictionary containing a link to the uploaded image under the key
            ``img_src``.

        :raises: ``prawcore.TooLarge`` if the overall request body is too large.
        :raises: :class:`.RedditAPIException` if there are other issues with the
            uploaded image. Unfortunately the exception info might not be very specific,
            so try through the website with the same image to see what the problem
            actually might be.

        For example:

        .. code-block:: python

            from praw.models import StylesheetImage

            reddit.subreddit("test").stylesheet.upload_header(StylesheetImage("header.png"))

        """
        return media._upload(self.subreddit, upload_type="header")

    def upload_mobile_banner(self, media: models.StylesheetAsset, /) -> None:
        """Upload an image for the :class:`.Subreddit`'s (redesign) mobile banner.

        :param media: The :class:`.StylesheetAsset` to upload.

        For example:

        .. code-block:: python

            from praw.models import StylesheetAsset

            subreddit = reddit.subreddit("test")
            subreddit.stylesheet.upload_mobile_banner(StylesheetAsset("banner.png"))

        Fails if the :class:`.Subreddit` does not have an additional image defined.

        :raises: ``prawcore.TooLarge`` if the overall request body is too large.
        :raises: :class:`.RedditAPIException` if there are other issues with the
            uploaded image. Unfortunately the exception info might not be very specific,
            so try through the website with the same image to see what the problem
            actually might be.

        """
        image_type = "mobileBannerImage"
        image_url = media._upload(self.subreddit, image_type=image_type)
        self._update_structured_styles({image_type: image_url})

    def upload_mobile_header(self, media: models.StylesheetImage, /) -> dict[str, str]:
        """Upload an image to be used as the :class:`.Subreddit`'s mobile header.

        :param media: The :class:`.StylesheetImage` to upload.

        :returns: A dictionary containing a link to the uploaded image under the key
            ``img_src``.

        :raises: ``prawcore.TooLarge`` if the overall request body is too large.
        :raises: :class:`.RedditAPIException` if there are other issues with the
            uploaded image. Unfortunately the exception info might not be very specific,
            so try through the website with the same image to see what the problem
            actually might be.

        For example:

        .. code-block:: python

            from praw.models import StylesheetImage

            reddit.subreddit("test").stylesheet.upload_mobile_header(StylesheetImage("header.png"))

        """
        return media._upload(self.subreddit, upload_type="banner")

    def upload_mobile_icon(self, media: models.StylesheetImage, /) -> dict[str, str]:
        """Upload an image to be used as the :class:`.Subreddit`'s mobile icon.

        :param media: The :class:`.StylesheetImage` to upload.

        :returns: A dictionary containing a link to the uploaded image under the key
            ``img_src``.

        :raises: ``prawcore.TooLarge`` if the overall request body is too large.
        :raises: :class:`.RedditAPIException` if there are other issues with the
            uploaded image. Unfortunately the exception info might not be very specific,
            so try through the website with the same image to see what the problem
            actually might be.

        For example:

        .. code-block:: python

            from praw.models import StylesheetImage

            reddit.subreddit("test").stylesheet.upload_mobile_icon(StylesheetImage("icon.png"))

        """
        return media._upload(self.subreddit, upload_type="icon")



"""Provide the Announcement class."""

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any

from praw.const import API_PATH
from praw.models.reddit.base import RedditBase
from praw.models.reddit.mixins.fullname import FullnameMixin

if TYPE_CHECKING:
    import praw


class Announcement(FullnameMixin, RedditBase):
    """A class that represents a Reddit announcement.

    .. include:: ../../typical_attributes.rst

    ================= =================================================================
    Attribute         Description
    ================= =================================================================
    ``body_html``     The announcement's body, rendered as HTML.
    ``body``          The announcement's body, in Markdown.
    ``id``            The fullname of the announcement, e.g. ``'ann_4sc833'``.
    ``permalink``     The announcement's permalink (to view on the web).
    ``read_at``       Timestamp the announcement was read, in ISO-8601 format. ``None``
                      if the announcement has not been read.
    ``read_datetime`` ``read_at`` as a timezone-aware :class:`datetime.datetime`, or
                      ``None`` if the announcement has not been read.
    ``sent_at``       Timestamp the announcement was sent, in ISO-8601 format.
    ``sent_datetime`` ``sent_at`` as a timezone-aware :class:`datetime.datetime`.
    ``subject``       The subject line of the announcement.
    ================= =================================================================

    """

    STR_FIELD = "id"

    @staticmethod
    def _parse_iso8601(value: str) -> datetime:
        # ``datetime.fromisoformat`` only accepts the ``Z`` suffix on Python 3.11+.
        return datetime.fromisoformat(value.replace("Z", "+00:00")).astimezone()

    @property
    def _kind(self) -> str:
        """The object's kind shortcode."""
        return "ann"

    @property
    def read_datetime(self) -> datetime | None:
        """The time the announcement was read as a timezone-aware :class:`datetime.datetime`.

        Returns ``None`` if the announcement has not been read. The returned object is
        localized to the system's timezone.

        """
        if self.read_at is None:
            return None
        return self._parse_iso8601(self.read_at)

    @property
    def sent_datetime(self) -> datetime:
        """The time the announcement was sent as a timezone-aware :class:`datetime.datetime`.

        The returned object is localized to the system's timezone.

        """
        return self._parse_iso8601(self.sent_at)

    def __init__(
        self,
        reddit: praw.Reddit,
        id: str | None = None,
        _data: dict[str, Any] | None = None,
    ) -> None:
        """Initialize an :class:`.Announcement` instance."""
        if (id, _data).count(None) != 1:
            msg = "Exactly one of 'id' or '_data' must be provided."
            raise TypeError(msg)
        if id:
            self.id = id
        super().__init__(reddit, _data=_data, _fetched=_data is not None)

    def hide(self) -> None:
        """Hide the announcement.

        Example usage:

        .. code-block:: python

            for announcement in reddit.announcements():
                announcement.hide()

        .. seealso::

            :meth:`.AnnouncementHelper.hide` to hide multiple announcements at once.

        """
        self._reddit.post(API_PATH["hide_announcements"], data={"ids": self.fullname})

    def mark_read(self) -> None:
        """Mark the announcement as read.

        Example usage:

        .. code-block:: python

            for announcement in reddit.announcements():
                if announcement.read_at is None:
                    announcement.mark_read()

        .. seealso::

            - :meth:`.AnnouncementHelper.mark_read` to mark multiple announcements as
              read at once.
            - :meth:`.AnnouncementHelper.mark_all_read` to mark all announcements as
              read.

        """
        self._reddit.post(API_PATH["read_announcements"], data={"ids": self.fullname})



"""Provide the InboxToggleableMixin class."""

from __future__ import annotations

from typing import TYPE_CHECKING

from praw.const import API_PATH

if TYPE_CHECKING:
    import praw


class InboxToggleableMixin:
    """Interface for classes that can optionally receive inbox replies."""

    if TYPE_CHECKING:
        # Provided by the host class (:class:`.RedditBase`).
        _reddit: praw.Reddit

        @property
        def fullname(self) -> str: ...  # noqa: D102

    def disable_inbox_replies(self) -> None:
        """Disable inbox replies for the item.

        .. note::

            This can only apply to items created by the authenticated user.

        Example usage:

        .. code-block:: python

            comment = reddit.comment("dkk4qjd")
            comment.disable_inbox_replies()

            submission = reddit.submission("8dmv8z")
            submission.disable_inbox_replies()

        .. seealso::

            :meth:`.enable_inbox_replies`

        """
        self._reddit.post(API_PATH["sendreplies"], data={"id": self.fullname, "state": False})

    def enable_inbox_replies(self) -> None:
        """Enable inbox replies for the item.

        .. note::

            This can only apply to items created by the authenticated user.

        Example usage:

        .. code-block:: python

            comment = reddit.comment("dkk4qjd")
            comment.enable_inbox_replies()

            submission = reddit.submission("8dmv8z")
            submission.enable_inbox_replies()

        .. seealso::

            :meth:`.disable_inbox_replies`

        """
        self._reddit.post(API_PATH["sendreplies"], data={"id": self.fullname, "state": True})



"""Provide the Subreddit class."""

from __future__ import annotations

from json import dumps, loads
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast, overload
from urllib.parse import urljoin

import websocket
from prawcore import Redirect

from praw.const import API_PATH
from praw.exceptions import MediaPostFailed, RedditAPIException, WebSocketException
from praw.models.listing.generator import ListingGenerator
from praw.models.listing.mixins import SubredditListingMixin
from praw.models.media import PostMedia
from praw.models.reddit.base import RedditBase
from praw.models.reddit.emoji import SubredditEmoji
from praw.models.reddit.mixins import CreatedMixin, FullnameMixin, MessageableMixin
from praw.models.reddit.rules import SubredditRules
from praw.models.reddit.subreddit.filters import SubredditFilters
from praw.models.reddit.subreddit.flair import SubredditFlair
from praw.models.reddit.subreddit.moderation import SubredditModeration
from praw.models.reddit.subreddit.modmail import Modmail
from praw.models.reddit.subreddit.quarantine import SubredditQuarantine
from praw.models.reddit.subreddit.relationship import (
    ContributorRelationship,
    ModeratorRelationship,
    SubredditRelationship,
)
from praw.models.reddit.subreddit.stream import SubredditStream
from praw.models.reddit.subreddit.stylesheet import SubredditStylesheet
from praw.models.reddit.subreddit.wiki import SubredditWiki
from praw.models.reddit.widgets import SubredditWidgets, WidgetEncoder
from praw.util import cachedproperty

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence

    import praw
    from praw import models
    from praw.models.reddit.collections import SubredditCollections


class Subreddit(MessageableMixin, SubredditListingMixin, FullnameMixin, CreatedMixin, RedditBase):
    """A class for Subreddits.

    To obtain an instance of this class for r/test execute:

    .. code-block:: python

        subreddit = reddit.subreddit("test")

    While r/all is not a real subreddit, it can still be treated like one. The following
    outputs the titles of the 25 hottest submissions in r/all:

    .. code-block:: python

        for submission in reddit.subreddit("all").hot(limit=25):
            print(submission.title)

    Multiple subreddits can be combined with a ``+`` like so:

    .. code-block:: python

        for submission in reddit.subreddit("redditdev+learnpython").top(time_filter="all"):
            print(submission)

    Subreddits can be filtered from combined listings as follows.

    .. note::

        These filters are ignored by certain methods, including :attr:`.comments`, and
        :meth:`.SubredditStream.comments`.

    .. code-block:: python

        for submission in reddit.subreddit("all-redditdev").new():
            print(submission)

    .. include:: ../../typical_attributes.rst

    ========================= ==========================================================
    Attribute                 Description
    ========================= ==========================================================
    ``can_assign_link_flair`` Whether users can assign their own link flair.
    ``can_assign_user_flair`` Whether users can assign their own user flair.
    ``created_utc``           Time the subreddit was created, represented in `Unix
                              Time`_.
    ``description``           Subreddit description, in Markdown.
    ``description_html``      Subreddit description, in HTML.
    ``display_name``          Name of the subreddit.
    ``icon_img``              The URL of the subreddit icon image.
    ``id``                    ID of the subreddit.
    ``name``                  Fullname of the subreddit.
    ``over18``                Whether the subreddit is NSFW.
    ``public_description``    Description of the subreddit, shown in searches and on the
                              "You must be invited to visit this community" page (if
                              applicable).
    ``spoilers_enabled``      Whether the spoiler tag feature is enabled.
    ``subscribers``           Count of subscribers.
    ``user_is_banned``        Whether the authenticated user is banned.
    ``user_is_moderator``     Whether the authenticated user is a moderator.
    ``user_is_subscriber``    Whether the authenticated user is subscribed.
    ========================= ==========================================================

    .. note::

        Trying to retrieve attributes of quarantined or private subreddits will result
        in a 403 error. Trying to retrieve attributes of a banned subreddit will result
        in a 404 error.

    .. _unix time: https://en.wikipedia.org/wiki/Unix_time

    """

    MAX_CAPTION_LENGTH = 180
    MESSAGE_PREFIX = "#"
    STR_FIELD = "display_name"

    # Bound at import time by the submission and collections modules to avoid circular
    # imports.
    _submission_class: type[models.Submission]
    _subreddit_collections_class: type[SubredditCollections]

    last_updated: int

    @staticmethod
    def _create_or_update(
        *,
        _reddit: praw.Reddit,
        allow_images: bool | None = None,
        allow_post_crossposts: bool | None = None,
        allow_top: bool | None = None,
        collapse_deleted_comments: bool | None = None,
        comment_score_hide_mins: int | None = None,
        description: str | None = None,
        domain: str | None = None,
        exclude_banned_modqueue: bool | None = None,
        header_hover_text: str | None = None,
        hide_ads: bool | None = None,
        key_color: str | None = None,
        lang: str | None = None,
        link_type: str | None = None,
        name: str | None = None,
        over_18: bool | None = None,
        public_description: str | None = None,
        public_traffic: bool | None = None,
        show_media: bool | None = None,
        show_media_preview: bool | None = None,
        spam_comments: bool | None = None,
        spam_links: bool | None = None,
        spam_selfposts: bool | None = None,
        spoilers_enabled: bool | None = None,
        sr: str | None = None,
        submit_link_label: str | None = None,
        submit_text: str | None = None,
        submit_text_label: str | None = None,
        subreddit_type: str | None = None,
        suggested_comment_sort: str | None = None,
        title: str | None = None,
        wiki_edit_age: int | None = None,
        wiki_edit_karma: int | None = None,
        wikimode: str | None = None,
        **other_settings: Any,
    ) -> None:
        model = {
            "allow_images": allow_images,
            "allow_post_crossposts": allow_post_crossposts,
            "allow_top": allow_top,
            "collapse_deleted_comments": collapse_deleted_comments,
            "comment_score_hide_mins": comment_score_hide_mins,
            "description": description,
            "domain": domain,
            "exclude_banned_modqueue": exclude_banned_modqueue,
            "header-title": header_hover_text,  # Remap here - better name
            "hide_ads": hide_ads,
            "key_color": key_color,
            "lang": lang,
            "link_type": link_type,
            "name": name,
            "over_18": over_18,
            "public_description": public_description,
            "public_traffic": public_traffic,
            "show_media": show_media,
            "show_media_preview": show_media_preview,
            "spam_comments": spam_comments,
            "spam_links": spam_links,
            "spam_selfposts": spam_selfposts,
            "spoilers_enabled": spoilers_enabled,
            "sr": sr,
            "submit_link_label": submit_link_label,
            "submit_text": submit_text,
            "submit_text_label": submit_text_label,
            "suggested_comment_sort": suggested_comment_sort,
            "title": title,
            "type": subreddit_type,
            "wiki_edit_age": wiki_edit_age,
            "wiki_edit_karma": wiki_edit_karma,
            "wikimode": wikimode,
        }

        model.update(other_settings)

        _reddit.post(API_PATH["site_admin"], data=model)

    @staticmethod
    def _subreddit_list(
        *,
        other_subreddits: Sequence[str | models.Subreddit] | None,
        subreddit: models.Subreddit,
    ) -> str:
        if other_subreddits:
            return ",".join([str(subreddit)] + [str(x) for x in other_subreddits])
        return str(subreddit)

    @staticmethod
    def _validate_gallery(images: list[dict[str, str | PostMedia]]) -> None:
        for image in images:
            media = image.get("media")
            if not isinstance(media, PostMedia):
                msg = "'media' is required and must be a PostMedia instance."
                raise TypeError(msg)
            if not len(cast("str", image.get("caption", ""))) <= Subreddit.MAX_CAPTION_LENGTH:
                msg = "Caption must be 180 characters or less."
                raise TypeError(msg)

    @staticmethod
    def _validate_inline_media(inline_media: models.InlineMedia) -> None:
        if not isinstance(inline_media.media, PostMedia):
            msg = "'media' must be a PostMedia instance."
            raise TypeError(msg)

    @cachedproperty
    def banned(self) -> SubredditRelationship:
        """Provide an instance of :class:`.SubredditRelationship`.

        For example, to ban a user try:

        .. code-block:: python

            reddit.subreddit("test").banned.add("spez", ban_reason="...")

        To list the banned users along with any notes, try:

        .. code-block:: python

            for ban in reddit.subreddit("test").banned():
                print(f"{ban}: {ban.note}")

        """
        return SubredditRelationship(self, "banned")

    @cachedproperty
    def collections(self) -> SubredditCollections:
        r"""Provide an instance of :class:`.SubredditCollections`.

        To see the permalinks of all :class:`.Collection`\ s that belong to a subreddit,
        try:

        .. code-block:: python

            for collection in reddit.subreddit("test").collections:
                print(collection.permalink)

        To get a specific :class:`.Collection` by its UUID or permalink, use one of the
        following:

        .. code-block:: python

            collection = reddit.subreddit("test").collections("some_uuid")
            collection = reddit.subreddit("test").collections(
                permalink="https://reddit.com/r/SUBREDDIT/collection/some_uuid"
            )

        """
        return self._subreddit_collections_class(self._reddit, self)

    @cachedproperty
    def contributor(self) -> ContributorRelationship:
        """Provide an instance of :class:`.ContributorRelationship`.

        Contributors are also known as approved submitters.

        To add a contributor try:

        .. code-block:: python

            reddit.subreddit("test").contributor.add("spez")

        """
        return ContributorRelationship(self, "contributor")

    @cachedproperty
    def emoji(self) -> SubredditEmoji:
        """Provide an instance of :class:`.SubredditEmoji`.

        This attribute can be used to discover all emoji for a subreddit:

        .. code-block:: python

            for emoji in reddit.subreddit("test").emoji:
                print(emoji)

        A single emoji can be lazily retrieved via:

        .. code-block:: python

            reddit.subreddit("test").emoji["emoji_name"]

        .. note::

            Attempting to access attributes of a nonexistent emoji will result in a
            :class:`.ClientException`.

        """
        return SubredditEmoji(self)

    @cachedproperty
    def filters(self) -> SubredditFilters:
        """Provide an instance of :class:`.SubredditFilters`.

        For example, to add a filter, run:

        .. code-block:: python

            reddit.subreddit("all").filters.add("test")

        """
        return SubredditFilters(self)

    @cachedproperty
    def flair(self) -> SubredditFlair:
        """Provide an instance of :class:`.SubredditFlair`.

        Use this attribute for interacting with a :class:`.Subreddit`'s flair. For
        example, to list all the flair for a subreddit which you have the ``flair``
        moderator permission on try:

        .. code-block:: python

            for flair in reddit.subreddit("test").flair():
                print(flair)

        Flair templates can be interacted with through this attribute via:

        .. code-block:: python

            for template in reddit.subreddit("test").flair.templates:
                print(template)

        """
        return SubredditFlair(self)

    @cachedproperty
    def mod(self) -> SubredditModeration:
        """Provide an instance of :class:`.SubredditModeration`.

        For example, to accept a moderation invite from r/test:

        .. code-block:: python

            reddit.subreddit("test").mod.accept_invite()

        """
        return SubredditModeration(self)

    @cachedproperty
    def moderator(self) -> ModeratorRelationship:
        """Provide an instance of :class:`.ModeratorRelationship`.

        For example, to add a moderator try:

        .. code-block:: python

            reddit.subreddit("test").moderator.add("spez")

        To list the moderators along with their permissions try:

        .. code-block:: python

            for moderator in reddit.subreddit("test").moderator():
                print(f"{moderator}: {moderator.mod_permissions}")

        """
        return ModeratorRelationship(self, "moderator")

    @cachedproperty
    def modmail(self) -> Modmail:
        """Provide an instance of :class:`.Modmail`.

        For example, to send a new modmail from r/test to u/spez with the subject
        ``"test"`` along with a message body of ``"hello"``:

        .. code-block:: python

            reddit.subreddit("test").modmail.create(subject="test", body="hello", recipient="spez")

        """
        return Modmail(self)

    @cachedproperty
    def muted(self) -> SubredditRelationship:
        """Provide an instance of :class:`.SubredditRelationship`.

        For example, muted users can be iterated through like so:

        .. code-block:: python

            for mute in reddit.subreddit("test").muted():
                print(f"{mute}: {mute.date}")

        """
        return SubredditRelationship(self, "muted")

    @cachedproperty
    def quaran(self) -> SubredditQuarantine:
        """Provide an instance of :class:`.SubredditQuarantine`.

        This property is named ``quaran`` because ``quarantine`` is a subreddit
        attribute returned by Reddit to indicate whether or not a subreddit is
        quarantined.

        To opt-in into a quarantined subreddit:

        .. code-block:: python

            reddit.subreddit("test").quaran.opt_in()

        """
        return SubredditQuarantine(self)

    @cachedproperty
    def rules(self) -> SubredditRules:
        """Provide an instance of :class:`.SubredditRules`.

        Use this attribute for interacting with a :class:`.Subreddit`'s rules.

        For example, to list all the rules for a subreddit:

        .. code-block:: python

            for rule in reddit.subreddit("test").rules:
                print(rule)

        Moderators can also add rules to the subreddit. For example, to make a rule
        called ``"No spam"`` in r/test:

        .. code-block:: python

            reddit.subreddit("test").rules.mod.add(
                short_name="No spam", kind="all", description="Do not spam. Spam bad"
            )

        """
        return SubredditRules(self)

    @cachedproperty
    def stream(self) -> SubredditStream:
        """Provide an instance of :class:`.SubredditStream`.

        Streams can be used to indefinitely retrieve new comments made to a subreddit,
        like:

        .. code-block:: python

            for comment in reddit.subreddit("test").stream.comments():
                print(comment)

        Additionally, new submissions can be retrieved via the stream. In the following
        example all submissions are fetched via the special r/all:

        .. code-block:: python

            for submission in reddit.subreddit("all").stream.submissions():
                print(submission)

        """
        return SubredditStream(self)

    @cachedproperty
    def stylesheet(self) -> SubredditStylesheet:
        """Provide an instance of :class:`.SubredditStylesheet`.

        For example, to add the css data ``.test{color:blue}`` to the existing
        stylesheet:

        .. code-block:: python

            subreddit = reddit.subreddit("test")
            stylesheet = subreddit.stylesheet()
            stylesheet.stylesheet += ".test{color:blue}"
            subreddit.stylesheet.update(stylesheet.stylesheet)

        """
        return SubredditStylesheet(self)

    @cachedproperty
    def widgets(self) -> models.SubredditWidgets:
        """Provide an instance of :class:`.SubredditWidgets`.

        **Example usage**

        Get all sidebar widgets:

        .. code-block:: python

            for widget in reddit.subreddit("test").widgets.sidebar:
                print(widget)

        Get ID card widget:

        .. code-block:: python

            print(reddit.subreddit("test").widgets.id_card)

        """
        return SubredditWidgets(self)

    @cachedproperty
    def wiki(self) -> SubredditWiki:
        """Provide an instance of :class:`.SubredditWiki`.

        This attribute can be used to discover all wikipages for a subreddit:

        .. code-block:: python

            for wikipage in reddit.subreddit("test").wiki:
                print(wikipage)

        To fetch the content for a given wikipage try:

        .. code-block:: python

            wikipage = reddit.subreddit("test").wiki["proof"]
            print(wikipage.content_md)

        """
        return SubredditWiki(self)

    @property
    def _kind(self) -> str:
        """The class's kind."""
        return self._reddit.config.kinds["subreddit"]

    def __init__(
        self,
        reddit: praw.Reddit,
        display_name: str | None = None,
        _data: dict[str, Any] | None = None,
    ) -> None:
        """Initialize a :class:`.Subreddit` instance.

        :param reddit: An instance of :class:`.Reddit`.
        :param display_name: The name of the subreddit.

        .. note::

            This class should not be initialized directly. Instead, obtain an instance
            via:

            .. code-block:: python

                subreddit = reddit.subreddit("test")

        """
        if (display_name, _data).count(None) != 1:
            msg = "Either 'display_name' or '_data' must be provided."
            raise TypeError(msg)
        if display_name:
            self.display_name = display_name
        super().__init__(reddit, _data=_data)
        self._path = API_PATH["subreddit"].format(subreddit=self)

    def _convert_to_fancypants(self, markdown_text: str) -> dict:
        """Convert a Markdown string to a dict for use with the ``richtext_json`` param.

        :param markdown_text: A Markdown string to convert.

        :returns: A dict in ``richtext_json`` format.

        """
        text_data = {"markdown_text": markdown_text, "output_mode": "rtjson"}
        return self._reddit.post(API_PATH["convert_rte_body"], data=text_data)["output"]

    def _fetch(self) -> None:
        data = self._fetch_data()
        data = data["data"]
        other = type(self)(self._reddit, _data=data)
        self.__dict__.update(other.__dict__)
        super()._fetch()

    def _fetch_info(self) -> tuple[str, dict[str, RedditBase], None]:
        return "subreddit_about", {"subreddit": self}, None

    def _submit_media(
        self, *, data: dict[Any, Any], timeout: int, without_websockets: bool
    ) -> models.Submission | None:
        """Submit and return an ``image``, ``video``, or ``videogif``.

        This is a helper method for submitting posts that are not link posts or self
        posts.

        """
        response = self._reddit.post(API_PATH["submit"], data=data)
        websocket_url = response["json"]["data"]["websocket_url"]
        connection = None
        if websocket_url is not None and not without_websockets:
            try:
                connection = websocket.create_connection(websocket_url, timeout=timeout)
            except (
                OSError,
                websocket.WebSocketException,
                BlockingIOError,
            ):
                msg = "Error establishing websocket connection."
                raise WebSocketException(msg) from None

        if connection is None:
            return None

        try:
            ws_update = loads(connection.recv())
            connection.close()
        except (OSError, websocket.WebSocketException, BlockingIOError):
            msg = "Websocket error. Check your media file. Your post may still have been created."
            raise WebSocketException(
                msg,
            ) from None
        if ws_update.get("type") == "failed":
            raise MediaPostFailed
        url = ws_update["payload"]["redirect"]
        return self._reddit.submission(url=url)

    def _upload_inline_media(self, inline_media: models.InlineMedia) -> models.InlineMedia:
        """Upload media for use in self posts and return ``inline_media``.

        :param inline_media: An :class:`.InlineMedia` object to validate and upload.

        """
        self._validate_inline_media(inline_media)
        inline_media.media_id = inline_media.media._upload(self._reddit, upload_type="selfpost")
        return inline_media

    def post_requirements(self) -> dict[str, str | int | bool]:
        """Get the post requirements for a subreddit.

        :returns: A dict with the various requirements.

        The returned dict contains the following keys:

        - ``domain_blacklist``
        - ``body_restriction_policy``
        - ``domain_whitelist``
        - ``title_regexes``
        - ``body_blacklisted_strings``
        - ``body_required_strings``
        - ``title_text_min_length``
        - ``is_flair_required``
        - ``title_text_max_length``
        - ``body_regexes``
        - ``link_repost_age``
        - ``body_text_min_length``
        - ``link_restriction_policy``
        - ``body_text_max_length``
        - ``title_required_strings``
        - ``title_blacklisted_strings``
        - ``guidelines_text``
        - ``guidelines_display_policy``

        For example, to fetch the post requirements for r/test:

        .. code-block:: python

            print(reddit.subreddit("test").post_requirements)

        """
        return self._reddit.get(API_PATH["post_requirements"].format(subreddit=str(self)))

    def search(
        self,
        query: str,
        *,
        sort: str = "relevance",
        syntax: str = "lucene",
        time_filter: str = "all",
        **generator_kwargs: Any,
    ) -> Iterator[models.Submission]:
        """Return a :class:`.ListingGenerator` for items that match ``query``.

        :param query: The query string to search for.
        :param sort: Can be one of: ``"relevance"``, ``"hot"``, ``"top"``, ``"new"``, or
            ``"comments"``. (default: ``"relevance"``).
        :param syntax: Can be one of: ``"cloudsearch"``, ``"lucene"``, or ``"plain"``
            (default: ``"lucene"``).
        :param time_filter: Can be one of: ``"all"``, ``"day"``, ``"hour"``,
            ``"month"``, ``"week"``, or ``"year"`` (default: ``"all"``).

        For more information on building a search query see:
        https://www.reddit.com/r/reddit.com/wiki/search/

        For example, to search all subreddits for ``"praw"`` try:

        .. code-block:: python

            for submission in reddit.subreddit("all").search("praw"):
                print(submission.title)

        """
        self._validate_time_filter(time_filter)
        not_all = self.display_name.lower() != "all"
        self._safely_add_arguments(
            arguments=generator_kwargs,
            key="params",
            q=query,
            restrict_sr=not_all,
            sort=sort,
            syntax=syntax,
            t=time_filter,
        )
        url = API_PATH["search"].format(subreddit=self)
        return ListingGenerator(self._reddit, url, **generator_kwargs)

    def sticky(self, *, number: int = 1) -> models.Submission:
        """Return a :class:`.Submission` object for a sticky of the subreddit.

        :param number: Specify which sticky to return. 1 appears at the top (default:
            ``1``).

        :raises: ``prawcore.NotFound`` if the sticky does not exist.

        For example, to get the stickied post on r/test:

        .. code-block:: python

            reddit.subreddit("test").sticky()

        """
        url = API_PATH["about_sticky"].format(subreddit=self)
        path = url
        try:
            self._reddit.get(url, params={"num": number})
        except Redirect as redirect:
            path = redirect.path
        reddit_url = self._reddit.config.reddit_url
        assert reddit_url is not None
        return self._submission_class(self._reddit, url=urljoin(reddit_url, path))

    @overload
    def submit(
        self,
        title: str,
        *,
        collection_id: str | None = ...,
        discussion_type: str | None = ...,
        draft_id: str | None = ...,
        flair_id: str | None = ...,
        flair_text: str | None = ...,
        inline_media: dict[str, models.InlineMedia] | None = ...,
        nsfw: bool = ...,
        resubmit: bool = ...,
        selftext: str,
        send_replies: bool = ...,
        spoiler: bool = ...,
    ) -> models.Submission: ...

    @overload
    def submit(
        self,
        title: str,
        *,
        collection_id: str | None = ...,
        discussion_type: str | None = ...,
        draft_id: str | None = ...,
        flair_id: str | None = ...,
        flair_text: str | None = ...,
        nsfw: bool = ...,
        resubmit: bool = ...,
        selftext: str | None = ...,
        send_replies: bool = ...,
        spoiler: bool = ...,
        url: str,
    ) -> models.Submission: ...

    @overload
    def submit(
        self,
        title: str,
        *,
        collection_id: str | None = ...,
        discussion_type: str | None = ...,
        flair_id: str | None = ...,
        flair_text: str | None = ...,
        image: models.PostMedia,
        nsfw: bool = ...,
        resubmit: bool = ...,
        selftext: str | None = ...,
        send_replies: bool = ...,
        spoiler: bool = ...,
        timeout: int = ...,
        without_websockets: bool = ...,
    ) -> models.Submission | None: ...

    @overload
    def submit(
        self,
        title: str,
        *,
        collection_id: str | None = ...,
        discussion_type: str | None = ...,
        flair_id: str | None = ...,
        flair_text: str | None = ...,
        gallery: list[models.PostMedia | dict[str, str | models.PostMedia]],
        nsfw: bool = ...,
        selftext: str | None = ...,
        send_replies: bool = ...,
        spoiler: bool = ...,
    ) -> models.Submission: ...

    @overload
    def submit(
        self,
        title: str,
        *,
        collection_id: str | None = ...,
        discussion_type: str | None = ...,
        flair_id: str | None = ...,
        flair_text: str | None = ...,
        nsfw: bool = ...,
        poll: dict[str, int | list[str]],
        resubmit: bool = ...,
        selftext: str | None = ...,
        send_replies: bool = ...,
        spoiler: bool = ...,
    ) -> models.Submission: ...

    @overload
    def submit(
        self,
        title: str,
        *,
        collection_id: str | None = ...,
        discussion_type: str | None = ...,
        flair_id: str | None = ...,
        flair_text: str | None = ...,
        nsfw: bool = ...,
        resubmit: bool = ...,
        selftext: str | None = ...,
        send_replies: bool = ...,
        spoiler: bool = ...,
        timeout: int = ...,
        video: models.PostMedia | dict[str, bool | models.PostMedia],
        without_websockets: bool = ...,
    ) -> models.Submission | None: ...

    def submit(
        self,
        title: str,
        *,
        collection_id: str | None = None,
        discussion_type: str | None = None,
        draft_id: str | None = None,
        flair_id: str | None = None,
        flair_text: str | None = None,
        gallery: list[models.PostMedia | dict[str, str | models.PostMedia]] | None = None,
        image: models.PostMedia | None = None,
        inline_media: dict[str, models.InlineMedia] | None = None,
        nsfw: bool = False,
        poll: dict[str, int | list[str]] | None = None,
        resubmit: bool = True,
        selftext: str | None = None,
        send_replies: bool = True,
        spoiler: bool = False,
        timeout: int = 10,
        url: str | None = None,
        video: models.PostMedia | dict[str, bool | models.PostMedia] | None = None,
        without_websockets: bool = False,
    ) -> models.Submission | None:
        r"""Add a submission to the :class:`.Subreddit`.

        :param title: The title of the submission.
        :param collection_id: The UUID of a :class:`.Collection` to add the
            newly-submitted post to.
        :param discussion_type: Set to ``"CHAT"`` to enable live discussion instead of
            traditional comments (default: ``None``).
        :param draft_id: The ID of a draft to submit. Only applies to text and link
            submissions.
        :param flair_id: The flair template to select (default: ``None``).
        :param flair_text: If the template's ``flair_text_editable`` value is ``True``,
            this value will set a custom text (default: ``None``). ``flair_id`` is
            required when ``flair_text`` is provided.
        :param gallery: A list of images to post as a gallery. Each item is either a
            :class:`.PostMedia` or a ``dict`` with the structure ``{"media":
            PostMedia("path"), "caption": "caption", "outbound_url": "url"}``, where
            only ``media`` is required.
        :param image: The :class:`.PostMedia` image to upload and post.
        :param inline_media: A dict of :class:`.InlineMedia` objects where the key is
            the placeholder name in ``selftext``. Only supported for text submissions.
        :param nsfw: Whether the submission should be marked NSFW (default: ``False``).
        :param poll: A ``dict`` with the structure ``{"duration": 3, "options": ["Yes",
            "No"]}``, where ``duration`` is the number of days the poll should accept
            votes (between ``1`` and ``7``, inclusive) and ``options`` is a list of two
            to six poll options as ``str``. Both keys are required.
        :param resubmit: When ``False``, an error will occur if the URL has already been
            submitted (default: ``True``).
        :param selftext: The Markdown formatted content for a ``text`` submission or
            optional Markdown-formatted body text for any other kind of submission. Use
            an empty string, ``""``, to make a title-only submission.
        :param send_replies: When ``True``, messages will be sent to the submission
            author when comments are made to the submission (default: ``True``).
        :param spoiler: Whether the submission should be marked as a spoiler (default:
            ``False``).
        :param timeout: Specifies a particular timeout, in seconds, for the WebSockets
            connection used by ``image`` and ``video`` submissions. Use to avoid
            "Websocket error" exceptions (default: ``10``).
        :param url: The URL for a ``link`` submission.
        :param video: The video to upload and post. Either a :class:`.PostMedia` or a
            ``dict`` with the structure ``{"media": PostMedia("path"), "gif": True,
            "thumbnail": PostMedia("path")}``, where only ``media`` is required. Set
            ``"gif"`` to ``True`` to submit the video as a videogif, which is
            essentially a silent video (default: ``False``). When ``"thumbnail"`` is not
            provided, the PRAW logo will be used as the thumbnail.
        :param without_websockets: Set to ``True`` to disable use of WebSockets for
            ``image`` and ``video`` submissions (see note below for an explanation). If
            ``True``, this method doesn't return anything (default: ``False``).

        :returns: A :class:`.Submission` object for the newly created submission, unless
            ``without_websockets`` is ``True`` for an ``image`` or ``video`` submission.

        :raises: :class:`.ClientException` if ``image`` or a ``gallery`` item's
            ``media`` refers to a file that is not an image, or if the ``video`` (or its
            ``media``) refers to a file that is not a video.

        At least one of ``gallery``, ``image``, ``poll``, ``selftext``, ``url``, or
        ``video`` must be provided. ``gallery``, ``image``, ``poll``, ``url``, and
        ``video`` are mutually exclusive, while ``selftext`` may accompany any of them
        as optional Markdown-formatted body text. ``selftext`` that accompanies another
        kind of submission does not support ``inline_media``.

        For example, to submit a URL to r/test do:

        .. code-block:: python

            title = "PRAW documentation"
            url = "https://praw.readthedocs.io"
            reddit.subreddit("test").submit(title, url=url)

        To submit an image to r/test do:

        .. code-block:: python

            from praw.models import PostMedia

            title = "My favorite picture"
            image = PostMedia("/path/to/image.png")
            reddit.subreddit("test").submit(title, image=image)

        To submit an image gallery to r/test do:

        .. code-block:: python

            from praw.models import PostMedia

            title = "My favorite pictures"
            gallery = [
                PostMedia("/path/to/image.png"),
                {
                    "media": PostMedia("/path/to/image2.png"),
                    "caption": "Image caption 2",
                },
                {
                    "media": PostMedia("/path/to/image3.png"),
                    "caption": "Image caption 3",
                    "outbound_url": "https://example.com/link3",
                },
            ]
            reddit.subreddit("test").submit(title, gallery=gallery)

        To submit a video to r/test do:

        .. code-block:: python

            from praw.models import PostMedia

            title = "My favorite movie"
            video = PostMedia("/path/to/video.mp4")
            reddit.subreddit("test").submit(title, video=video)

        To submit a videogif with a custom thumbnail instead, do:

        .. code-block:: python

            from praw.models import PostMedia

            title = "My favorite gif"
            video = {
                "gif": True,
                "media": PostMedia("/path/to/video.mp4"),
                "thumbnail": PostMedia("/path/to/thumbnail.png"),
            }
            reddit.subreddit("test").submit(title, video=video)

        To submit a poll to r/test do:

        .. code-block:: python

            title = "Do you like PRAW?"
            poll = {"duration": 3, "options": ["Yes", "No"]}
            reddit.subreddit("test").submit(title, poll=poll)

        To submit a self post with inline media do:

        .. code-block:: python

            from praw.models import InlineGif, InlineImage, InlineVideo, PostMedia

            gif = InlineGif(caption="optional caption", media=PostMedia("path/to/image.gif"))
            image = InlineImage(caption="optional caption", media=PostMedia("path/to/image.jpg"))
            video = InlineVideo(caption="optional caption", media=PostMedia("path/to/video.mp4"))
            selftext = "Text with a gif {gif1} an image {image1} and a video {video1} inline"
            media = {"gif1": gif, "image1": image, "video1": video}
            reddit.subreddit("test").submit("title", inline_media=media, selftext=selftext)

        .. note::

            Inserted media will have a padding of ``\\n\\n`` automatically added. This
            is due to the weirdness with Reddit's API. Using the example above, the
            result selftext body will look like so:

            .. code-block::

                Text with a gif

                ![gif](u1rchuphryq51 "optional caption")

                an image

                ![img](srnr8tshryq51 "optional caption")

                and video

                ![video](gmc7rvthryq51 "optional caption")

                inline

        .. note::

            For ``image`` and ``video`` submissions, Reddit's API uses WebSockets to
            respond with the link of the newly created post. If this fails, the method
            will raise :class:`.WebSocketException`. Occasionally, the Reddit post will
            still be created. More often, there is an error with the media file. If you
            frequently get exceptions but successfully created posts, try setting the
            ``timeout`` parameter to a value above 10.

            To disable the use of WebSockets, set ``without_websockets=True``. This will
            make the method return ``None``, though the post will still be created. You
            may wish to do this if you are running your program in a restricted network
            environment, or using a proxy that doesn't support WebSockets connections.

        .. note::

            To submit a post to a subreddit with the ``"news"`` flair, you can get the
            flair id like this:

            .. code-block::

                choices = list(subreddit.flair.link_templates.user_selectable())
                template_id = next(x for x in choices if x["flair_text"] == "news")["flair_template_id"]
                subreddit.submit("title", flair_id=template_id, url="https://www.news.com/")

        """
        provided = [
            name
            for name, value in (
                ("gallery", gallery),
                ("image", image),
                ("poll", poll),
                ("url", url),
                ("video", video),
            )
            if value is not None
        ]
        if len(provided) > 1:
            msg = f"Only one of 'gallery', 'image', 'poll', 'url', or 'video' can be provided ({', '.join(repr(name) for name in provided)} given)."
            raise TypeError(msg)
        kind = provided[0] if provided else None
        # test for empty string in selftext for title-only submissions
        if kind is None and not (bool(selftext) or selftext == ""):  # noqa: PLC1901
            msg = "At least one of 'gallery', 'image', 'poll', 'selftext', 'url', or 'video' must be provided."
            raise TypeError(msg)
        if inline_media and kind is not None:
            msg = f"'inline_media' is only supported for text submissions. Only Markdown text can be used for the selftext of a {kind!r} submission."
            raise TypeError(msg)

        data = {
            "nsfw": bool(nsfw),
            "sendreplies": bool(send_replies),
            "spoiler": bool(spoiler),
            "sr": str(self),
            "title": title,
            "validate_on_submit": True,
        }
        data.update({
            key: value
            for key, value in (
                ("collection_id", collection_id),
                ("discussion_type", discussion_type),
                ("flair_id", flair_id),
                ("flair_text", flair_text),
            )
            if value is not None
        })

        if gallery is not None:
            images: list[dict[str, str | PostMedia]] = [
                {"media": item} if isinstance(item, PostMedia) else item for item in gallery
            ]
            self._validate_gallery(images)
            data.update(api_type="json", items=[], show_error_list=True)
            if selftext is not None:
                data["text"] = selftext
            for image_item in images:
                data["items"].append({
                    "caption": image_item.get("caption", ""),
                    "media_id": cast("PostMedia", image_item["media"])._upload(
                        self._reddit,
                        expected_mime_prefix="image",
                        upload_type="gallery",
                    ),
                    "outbound_url": image_item.get("outbound_url", ""),
                })
            response = self._reddit.request(json=data, method="POST", path=API_PATH["submit_gallery_post"])["json"]
            if response["errors"]:
                raise RedditAPIException(response["errors"])
            return self._reddit.submission(url=response["data"]["url"])

        data["resubmit"] = bool(resubmit)

        if poll is not None:
            invalid_keys = poll.keys() - {"duration", "options"}
            if invalid_keys:
                msg = f"'poll' contains invalid keys: {', '.join(repr(key) for key in sorted(invalid_keys))}."
                raise TypeError(msg)
            missing_keys = {"duration", "options"} - poll.keys()
            if missing_keys:
                msg = f"'poll' is missing required keys: {', '.join(repr(key) for key in sorted(missing_keys))}."
                raise TypeError(msg)
            data.update(
                duration=poll["duration"],
                options=poll["options"],
                text=selftext if selftext is not None else "",
            )
            return self._reddit.post(API_PATH["submit_poll_post"], json=data)

        if image is not None:
            if selftext is not None:
                data["text"] = selftext
            data.update(kind="image", url=image._upload(self._reddit, expected_mime_prefix="image"))
            return self._submit_media(data=data, timeout=timeout, without_websockets=without_websockets)

        if video is not None:
            if isinstance(video, PostMedia):
                video = {"media": video}
            invalid_keys = video.keys() - {"gif", "media", "thumbnail"}
            if invalid_keys:
                msg = f"'video' contains invalid keys: {', '.join(repr(key) for key in sorted(invalid_keys))}."
                raise TypeError(msg)
            video_media = video.get("media")
            if not isinstance(video_media, PostMedia):
                msg = "'media' is required and must be a PostMedia instance."
                raise TypeError(msg)
            thumbnail_media = cast("PostMedia | None", video.get("thumbnail"))
            if thumbnail_media is None:
                # if we're uploading without a thumbnail, use the PRAW logo
                logo_path = Path(__file__).absolute().parent.parent.parent.parent / "images" / "PRAW logo.png"
                thumbnail_media = PostMedia(str(logo_path))
            if selftext is not None:
                data["text"] = selftext
            data.update(
                kind="videogif" if video.get("gif") else "video",
                url=video_media._upload(self._reddit, expected_mime_prefix="video"),
                video_poster_url=thumbnail_media._upload(self._reddit),
            )
            return self._submit_media(data=data, timeout=timeout, without_websockets=without_websockets)

        if draft_id is not None:
            data["draft_id"] = draft_id
        if url is not None:
            data.update(kind="link", url=url)
            # we can ignore an empty string for selftext here b/c body text is optional for link posts
            if selftext:
                data["text"] = selftext
        else:
            data.update(kind="self")
            if inline_media:
                assert selftext is not None
                body = selftext.format(**{
                    placeholder: self._upload_inline_media(media) for placeholder, media in inline_media.items()
                })
                data.update(richtext_json=dumps(self._convert_to_fancypants(body)))
            else:
                data.update(text=selftext)

        return self._reddit.post(API_PATH["submit"], data=data)

    def subscribe(self, *, other_subreddits: list[models.Subreddit] | None = None) -> None:
        """Subscribe to the subreddit.

        :param other_subreddits: When provided, also subscribe to the provided list of
            subreddits.

        For example, to subscribe to r/test:

        .. code-block:: python

            reddit.subreddit("test").subscribe()

        """
        data = {
            "action": "sub",
            "skip_inital_defaults": True,
            "sr_name": self._subreddit_list(other_subreddits=other_subreddits, subreddit=self),
        }
        self._reddit.post(API_PATH["subscribe"], data=data)

    def traffic(self) -> dict[str, list[list[int]]]:
        """Return a dictionary of the :class:`.Subreddit`'s traffic statistics.

        :raises: ``prawcore.NotFound`` when the traffic stats aren't available to the
            authenticated user, that is, they are not public and the authenticated user
            is not a moderator of the subreddit.

        The traffic method returns a dict with three keys. The keys are ``day``,
        ``hour`` and ``month``. Each key contains a list of lists with 3 or 4 values.
        The first value is a timestamp indicating the start of the category (start of
        the day for the ``day`` key, start of the hour for the ``hour`` key, etc.). The
        second, third, and fourth values indicate the unique pageviews, total pageviews,
        and subscribers, respectively.

        .. note::

            The ``hour`` key does not contain subscribers, and therefore each sub-list
            contains three values.

        For example, to get the traffic stats for r/test:

        .. code-block:: python

            stats = reddit.subreddit("test").traffic()

        """
        return self._reddit.get(API_PATH["about_traffic"].format(subreddit=self))

    def unsubscribe(self, *, other_subreddits: list[models.Subreddit] | None = None) -> None:
        """Unsubscribe from the subreddit.

        :param other_subreddits: When provided, also unsubscribe from the provided list
            of subreddits.

        To unsubscribe from r/test:

        .. code-block:: python

            reddit.subreddit("test").unsubscribe()

        """
        data = {
            "action": "unsub",
            "sr_name": self._subreddit_list(other_subreddits=other_subreddits, subreddit=self),
        }
        self._reddit.post(API_PATH["subscribe"], data=data)


WidgetEncoder._subreddit_class = Subreddit



"""Provide classes for uploading media to Reddit."""

from __future__ import annotations

import sys
from functools import cached_property
from pathlib import Path
from typing import IO, TYPE_CHECKING, Any, ClassVar, cast

from defusedxml import ElementTree
from prawcore.exceptions import ServerError

from praw.const import API_PATH, JPEG_HEADER
from praw.exceptions import (
    ClientException,
    RedditAPIException,
    RedditErrorItem,
    TooLargeMediaException,
)

if sys.version_info >= (3, 13):  # pragma: no cover
    from mimetypes import guess_file_type
else:  # pragma: no cover
    from mimetypes import guess_type as guess_file_type

if TYPE_CHECKING:
    from requests import Response

    import praw
    from praw import models


class Media:
    """Base class representing media that can be uploaded to Reddit.

    Use one of the subclasses, e.g., :class:`.EmojiMedia` or :class:`.PostMedia`,
    depending on what the media is being uploaded for.

    """

    LEASE_API_PATH: ClassVar[str]
    LEASE_RESPONSE_KEY: ClassVar[str] = "s3UploadLease"

    @staticmethod
    def _raise_upload_error(response: Response, /) -> None:
        raise ServerError(response)

    def __eq__(self, other: object) -> bool:
        """Return whether the other instance equals the current."""
        return type(other) is type(self) and self.name == other.name and self._data == other._data

    def __hash__(self) -> int:
        """Return the hash of the current instance."""
        return hash((self.__class__.__name__, self.name, self._data))

    def __init__(self, fp: str | bytes, /, name: str | None = None) -> None:
        """Initialize a :class:`.Media` instance.

        :param fp: The path to a media file as a string, or the content of a media file
            as a :class:`bytes` object.
        :param name: The name of the media file, e.g., ``"picture.png"``. The name is
            used to infer the media's MIME type. When ``fp`` is a path the name is
            derived from it, otherwise this parameter is required.

        """
        if isinstance(fp, bytes):
            if not name:
                msg = "'name' is required when 'fp' is a bytes object."
                raise ValueError(msg)
            data = fp
        else:
            path = Path(fp)
            name = name or path.name
            data = path.read_bytes()
        self._data = data
        self.name = name

    def __repr__(self) -> str:
        """Return a string representation of the instance."""
        return f"<{self.__class__.__name__} name={self.name!r}>"

    def _build_lease_data(self, **additional_data: str) -> dict[str, str]:
        return {"filepath": self.name, "mimetype": self._mime_type, **additional_data}

    def _lease_and_post(
        self, lease_url: str, reddit: praw.Reddit, /, **additional_lease_data: str
    ) -> tuple[dict[str, Any], dict[str, str], str]:
        lease_data = self._build_lease_data(**additional_lease_data)
        lease_response, upload_data, upload_url = self._obtain_lease(lease_data, lease_url, reddit)
        response = self._post_to_s3(reddit, upload_data, upload_url)
        if not response.ok:
            self._raise_upload_error(response)
        return lease_response, upload_data, upload_url

    @cached_property
    def _mime_type(self) -> str:
        mime_type, _ = guess_file_type(self.name)
        if mime_type is None:
            msg = f"Unable to determine the MIME type of {self.name!r}."
            raise ClientException(msg)
        return mime_type

    def _obtain_lease(
        self, lease_data: dict[str, str], lease_url: str, reddit: praw.Reddit, /
    ) -> tuple[dict[str, Any], dict[str, str], str]:
        lease_response = reddit.post(lease_url, data=lease_data)
        upload_lease = lease_response[self.LEASE_RESPONSE_KEY]
        upload_data = {item["name"]: item["value"] for item in upload_lease["fields"]}
        upload_url = f"https:{upload_lease['action']}"
        return lease_response, upload_data, upload_url

    def _post_to_s3(self, reddit: praw.Reddit, upload_data: dict[str, str], upload_url: str, /) -> Response:
        assert reddit._core is not None
        return reddit._core.requestor.request(
            "POST",
            upload_url,
            data=upload_data,
            files={"file": (self.name, self._data)},
        )

    def _upload(self, subreddit: models.Subreddit, /, **additional_lease_data: str) -> str:
        """Upload the media to Reddit.

        :param subreddit: The subreddit the media is associated with.
        :param additional_lease_data: Additional data to include in the upload lease
            request.

        :returns: The URL of the uploaded media.

        """
        lease_url = self.LEASE_API_PATH.format(subreddit=subreddit)
        _, upload_data, upload_url = self._lease_and_post(lease_url, subreddit._reddit, **additional_lease_data)
        return f"{upload_url}/{upload_data['key']}"


class EmojiMedia(Media):
    """Media to be uploaded as a subreddit emoji.

    See :meth:`.SubredditEmoji.add`.

    """

    LEASE_API_PATH = API_PATH["emoji_lease"]

    def _upload(self, subreddit: models.Subreddit, /, **additional_lease_data: str) -> str:
        """Upload the media to Reddit.

        :param subreddit: The subreddit the emoji is associated with.
        :param additional_lease_data: Additional data to include in the upload lease
            request.

        :returns: The S3 key of the uploaded media.

        """
        lease_url = self.LEASE_API_PATH.format(subreddit=subreddit)
        _, upload_data, _ = self._lease_and_post(lease_url, subreddit._reddit, **additional_lease_data)
        return upload_data["key"]


class PostMedia(Media):
    """Media to be uploaded as part of a submission.

    See :meth:`.Subreddit.submit` and :class:`.InlineMedia`.

    """

    LEASE_API_PATH = API_PATH["media_asset"]
    LEASE_RESPONSE_KEY = "args"

    @staticmethod
    def _parse_xml_response(response: Response, /) -> None:
        """Parse the XML from a response and raise any errors found."""
        root = ElementTree.fromstring(response.text)
        tags = [element.tag for element in root]
        if tags[:4] == ["Code", "Message", "ProposedSize", "MaxSizeAllowed"]:
            # Returned if media is too big
            *_, actual, maximum_size = (element.text for element in root[:4])
            raise TooLargeMediaException(actual=int(actual or 0), maximum_size=int(maximum_size or 0))

    @staticmethod
    def _raise_upload_error(response: Response, /) -> None:
        PostMedia._parse_xml_response(response)
        Media._raise_upload_error(response)

    def _upload(  # pyright: ignore[reportIncompatibleMethodOverride]  # post media is uploaded with a Reddit instance rather than a Subreddit
        self,
        reddit: praw.Reddit,
        /,
        *,
        expected_mime_prefix: str | None = None,
        upload_type: str = "link",
    ) -> str:
        """Upload the media to Reddit (undocumented endpoint).

        Unlike the other :class:`.Media` subclasses, post media is not associated with a
        subreddit when uploaded, so this method takes a :class:`.Reddit` instance.

        :param reddit: The :class:`.Reddit` instance to upload with.
        :param expected_mime_prefix: If provided, enforce that the media has a MIME type
            that starts with the provided prefix.
        :param upload_type: One of ``"link"``, ``"gallery"``, or ``"selfpost"``
            (default: ``"link"``).

        :returns: The URL of the uploaded media when ``upload_type`` is ``"link"``,
            otherwise the media's asset ID.

        """
        if expected_mime_prefix is not None and self._mime_type.partition("/")[0] != expected_mime_prefix:
            msg = f"Expected a mimetype starting with {expected_mime_prefix!r} but got mimetype {self._mime_type!r} (from file name {self.name!r})."
            raise ClientException(msg)
        lease_response, upload_data, upload_url = self._lease_and_post(self.LEASE_API_PATH, reddit)
        if upload_type == "link":
            return f"{upload_url}/{upload_data['key']}"
        return lease_response["asset"]["asset_id"]


class StylesheetAsset(Media):
    """Media to be uploaded as a subreddit style asset, e.g., a banner.

    See :meth:`.SubredditStylesheet.upload_banner`.

    """

    LEASE_API_PATH = API_PATH["style_asset_lease"]

    def _upload(  # pyright: ignore[reportIncompatibleMethodOverride]  # style assets require an image_type rather than arbitrary lease data
        self, subreddit: models.Subreddit, /, *, image_type: str
    ) -> str:
        """Upload the media to Reddit.

        :param subreddit: The subreddit the style asset is associated with.
        :param image_type: The type of style asset, e.g., ``"bannerBackgroundImage"``.

        :returns: The URL of the uploaded media.

        """
        return super()._upload(subreddit, imagetype=image_type)


class StylesheetImage(Media):
    """Media to be uploaded as a subreddit stylesheet image.

    See :meth:`.SubredditStylesheet.upload`.

    """

    UPLOAD_API_PATH = API_PATH["upload_image"]

    @cached_property
    def _image_type(self) -> str:
        return "jpg" if self._data.startswith(JPEG_HEADER) else "png"

    def _upload(  # pyright: ignore[reportIncompatibleMethodOverride]  # stylesheet image upload returns the raw response dict
        self, subreddit: models.Subreddit, /, **additional_data: str
    ) -> dict[str, Any]:
        """Upload the media to Reddit.

        :param subreddit: The subreddit the stylesheet image is associated with.
        :param additional_data: Additional data to include in the upload request.

        :returns: A dictionary containing a link to the uploaded image under the key
            ``img_src``.

        """
        data = {"img_type": self._image_type, **additional_data}
        url = self.UPLOAD_API_PATH.format(subreddit=subreddit)
        # ``requests`` accepts ``(filename, content)`` tuples for file uploads, but
        # ``Reddit.post`` types ``files`` as ``dict[str, IO]``.
        files = cast("dict[str, IO[Any]]", {"file": (self.name, self._data)})
        response = subreddit._reddit.post(url, data=data, files=files)
        if response["errors"]:
            error_type = response["errors"][0]
            error_value = response.get("errors_values", [""])[0]
            assert error_type in {
                "BAD_CSS_NAME",
                "IMAGE_ERROR",
            }, "Please file a bug with PRAW."
            raise RedditAPIException([RedditErrorItem(error_type=error_type, field="", message=error_value or "")])
        return response


class WidgetMedia(Media):
    """Media to be uploaded for use in a subreddit widget.

    See :meth:`.SubredditWidgetsModeration.upload_image`.

    """

    LEASE_API_PATH = API_PATH["widget_lease"]



"""Provide the InboxableMixin class."""

from __future__ import annotations

from typing import TYPE_CHECKING, cast

from praw.const import API_PATH

if TYPE_CHECKING:
    import praw
    from praw import models


class InboxableMixin:
    """Interface for :class:`.RedditBase` subclasses that originate from the inbox."""

    if TYPE_CHECKING:
        # Provided by the host class (:class:`.RedditBase`).
        _reddit: praw.Reddit

        @property
        def fullname(self) -> str: ...  # noqa: D102

    def block(self) -> None:
        """Block the user who sent the item.

        .. note::

            This method pertains only to objects which were retrieved via the inbox.

        Example usage:

        .. code-block:: python

            comment = reddit.comment("dkk4qjd")
            comment.block()

            # or, identically:

            comment.author.block()

        """
        self._reddit.post(API_PATH["block"], data={"id": self.fullname})

    def collapse(self) -> None:
        """Mark the item as collapsed.

        .. note::

            This method pertains only to objects which were retrieved via the inbox.

        Example usage:

        .. code-block:: python

            inbox = reddit.inbox()

            # select first inbox item and collapse it message = next(inbox)
            message.collapse()

        .. seealso::

            :meth:`.uncollapse`

        """
        self._reddit.inbox.collapse([cast("models.Message", self)])

    def mark_read(self) -> None:
        """Mark a single inbox item as read.

        .. note::

            This method pertains only to objects which were retrieved via the inbox.

        Example usage:

        .. code-block:: python

            inbox = reddit.inbox.unread()

            for message in inbox:
                # process unread messages
                ...

        .. seealso::

            :meth:`.mark_unread`

        To mark the whole inbox as read with a single network request, use
        :meth:`.Inbox.mark_all_read`

        """
        self._reddit.inbox.mark_read([cast("models.Comment | models.Message", self)])

    def mark_unread(self) -> None:
        """Mark the item as unread.

        .. note::

            This method pertains only to objects which were retrieved via the inbox.

        Example usage:

        .. code-block:: python

            inbox = reddit.inbox(limit=10)

            for message in inbox:
                # process messages
                ...

        .. seealso::

            :meth:`.mark_read`

        """
        self._reddit.inbox.mark_unread([cast("models.Comment | models.Message", self)])

    def uncollapse(self) -> None:
        """Mark the item as uncollapsed.

        .. note::

            This method pertains only to objects which were retrieved via the inbox.

        Example usage:

        .. code-block:: python

            inbox = reddit.inbox()

            # select first inbox item and uncollapse it
            message = next(inbox)
            message.uncollapse()

        .. seealso::

            :meth:`.collapse`

        """
        self._reddit.inbox.uncollapse([cast("models.Message", self)])



"""Provide the SubredditFlairTemplates class."""

from __future__ import annotations

from csv import writer
from io import StringIO
from typing import TYPE_CHECKING, Any, cast

from praw.const import API_PATH
from praw.exceptions import InvalidFlairTemplateID
from praw.models.listing.generator import ListingGenerator
from praw.models.reddit.base import RedditBase
from praw.util import cachedproperty

if TYPE_CHECKING:
    from collections.abc import Iterator

    from praw import models


class SubredditFlair:
    """Provide a set of functions to interact with a :class:`.Subreddit`'s flair."""

    @cachedproperty
    def link_templates(
        self,
    ) -> SubredditLinkFlairTemplates:
        """Provide an instance of :class:`.SubredditLinkFlairTemplates`.

        Use this attribute for interacting with a :class:`.Subreddit`'s link flair
        templates. For example to list all the link flair templates for a subreddit
        which you have the ``flair`` moderator permission on try:

        .. code-block:: python

            for template in reddit.subreddit("test").flair.link_templates:
                print(template)

        """
        return SubredditLinkFlairTemplates(self.subreddit)

    @cachedproperty
    def templates(
        self,
    ) -> SubredditRedditorFlairTemplates:
        """Provide an instance of :class:`.SubredditRedditorFlairTemplates`.

        Use this attribute for interacting with a :class:`.Subreddit`'s flair templates.
        For example to list all the flair templates for a subreddit which you have the
        ``flair`` moderator permission on try:

        .. code-block:: python

            for template in reddit.subreddit("test").flair.templates:
                print(template)

        """
        return SubredditRedditorFlairTemplates(self.subreddit)

    def __call__(
        self,
        redditor: models.Redditor | str | None = None,
        **generator_kwargs: Any,
    ) -> Iterator[models.Redditor]:
        """Return a :class:`.ListingGenerator` for Redditors and their flairs.

        :param redditor: When provided, yield at most a single :class:`.Redditor`
            instance (default: ``None``).

        Additional keyword arguments are passed in the initialization of
        :class:`.ListingGenerator`.

        Usage:

        .. code-block:: python

            for flair in reddit.subreddit("test").flair(limit=None):
                print(flair)

        """
        RedditBase._safely_add_arguments(arguments=generator_kwargs, key="params", name=redditor)
        generator_kwargs.setdefault("limit", None)
        url = API_PATH["flairlist"].format(subreddit=self.subreddit)
        return ListingGenerator(self.subreddit._reddit, url, **generator_kwargs)

    def __init__(self, subreddit: models.Subreddit) -> None:
        """Initialize a :class:`.SubredditFlair` instance.

        :param subreddit: The subreddit whose flair to work with.

        """
        self.subreddit = subreddit

    def configure(
        self,
        *,
        link_position: str = "left",
        link_self_assign: bool = False,
        position: str = "right",
        self_assign: bool = False,
        **settings: Any,
    ) -> None:
        """Update the :class:`.Subreddit`'s flair configuration.

        :param link_position: One of ``"left"``, ``"right"``, or ``False`` to disable
            (default: ``"left"``).
        :param link_self_assign: Permit self assignment of link flair (default:
            ``False``).
        :param position: One of ``"left"``, ``"right"``, or ``False`` to disable
            (default: ``"right"``).
        :param self_assign: Permit self assignment of user flair (default: ``False``).

        .. code-block:: python

            subreddit = reddit.subreddit("test")
            subreddit.flair.configure(link_position="right", self_assign=True)

        Additional keyword arguments can be provided to handle new settings as Reddit
        introduces them.

        """
        data = {
            "flair_enabled": bool(position),
            "flair_position": position or "right",
            "flair_self_assign_enabled": self_assign,
            "link_flair_position": link_position or "",
            "link_flair_self_assign_enabled": link_self_assign,
        }
        data.update(settings)
        url = API_PATH["flairconfig"].format(subreddit=self.subreddit)
        self.subreddit._reddit.post(url, data=data)

    def delete(self, redditor: models.Redditor | str) -> None:
        """Delete flair for a :class:`.Redditor`.

        :param redditor: A redditor name or :class:`.Redditor` instance.

        .. seealso::

            :meth:`~.SubredditFlair.update` to delete the flair of many redditors at
            once.

        """
        url = API_PATH["deleteflair"].format(subreddit=self.subreddit)
        self.subreddit._reddit.post(url, data={"name": str(redditor)})

    def delete_all(self) -> list[dict[str, str | bool | dict[str, str]]]:
        """Delete all :class:`.Redditor` flair in the :class:`.Subreddit`.

        :returns: List of dictionaries indicating the success or failure of each delete.

        """
        return self.update(cast("dict[str, str | models.Redditor]", x)["user"] for x in self())

    def set(
        self,
        redditor: models.Redditor | str,
        *,
        css_class: str = "",
        flair_template_id: str | None = None,
        text: str = "",
    ) -> None:
        """Set flair for a :class:`.Redditor`.

        :param redditor: A redditor name or :class:`.Redditor` instance.
        :param text: The flair text to associate with the :class:`.Redditor` or
            :class:`.Submission` (default: ``""``).
        :param css_class: The css class to associate with the flair html (default:
            ``""``). Use either this or ``flair_template_id``.
        :param flair_template_id: The ID of the flair template to be used (default:
            ``None``). Use either this or ``css_class``.

        This method can only be used by an authenticated user who is a moderator of the
        associated :class:`.Subreddit`.

        For example:

        .. code-block:: python

            reddit.subreddit("test").flair.set("bboe", text="PRAW author", css_class="mods")
            template = "6bd28436-1aa7-11e9-9902-0e05ab0fad46"
            reddit.subreddit("test").flair.set(
                "spez", text="Reddit CEO", flair_template_id=template
            )

        """
        if css_class and flair_template_id is not None:
            msg = "Parameter 'css_class' cannot be used in conjunction with 'flair_template_id'."
            raise TypeError(msg)
        data = {"name": str(redditor), "text": text}
        if flair_template_id is not None:
            data["flair_template_id"] = flair_template_id
            url = API_PATH["select_flair"].format(subreddit=self.subreddit)
        else:
            data["css_class"] = css_class
            url = API_PATH["flair"].format(subreddit=self.subreddit)
        self.subreddit._reddit.post(url, data=data)

    def update(
        self,
        flair_list: Iterator[str | models.Redditor | dict[str, str | models.Redditor]],
        *,
        css_class: str = "",
        text: str = "",
    ) -> list[dict[str, str | bool | dict[str, str]]]:
        """Set or clear the flair for many redditors at once.

        :param flair_list: Each item in this list should be either:

            - The name of a redditor.
            - An instance of :class:`.Redditor`.
            - A dictionary mapping keys ``"user"``, ``"flair_text"``, and
              ``"flair_css_class"`` to their respective values. The ``"user"`` key
              should map to a redditor, as described above. When a dictionary isn't
              provided, or the dictionary is missing either ``"flair_text"`` or
              ``"flair_css_class"`` keys, the default values will come from the other
              arguments.
        :param css_class: The css class to use when not explicitly provided in
            ``flair_list`` (default: ``""``).
        :param text: The flair text to use when not explicitly provided in
            ``flair_list`` (default: ``""``).

        :returns: List of dictionaries indicating the success or failure of each update.

        For example, to clear the flair text, and set the ``"praw"`` flair css class on
        a few users try:

        .. code-block:: python

            subreddit.flair.update(["bboe", "spez", "spladug"], css_class="praw")

        """
        temp_lines = StringIO()
        for item in flair_list:
            if isinstance(item, dict):
                writer(temp_lines).writerow([
                    str(item["user"]),
                    item.get("flair_text", text),
                    item.get("flair_css_class", css_class),
                ])
            else:
                writer(temp_lines).writerow([str(item), text, css_class])

        lines = temp_lines.getvalue().splitlines()
        temp_lines.close()
        response = []
        url = API_PATH["flaircsv"].format(subreddit=self.subreddit)
        while lines:
            data = {"flair_csv": "\n".join(lines[:100])}
            response.extend(self.subreddit._reddit.post(url, data=data))
            lines = lines[100:]
        return response


class SubredditFlairTemplates:
    """Provide functions to interact with a :class:`.Subreddit`'s flair templates."""

    @staticmethod
    def flair_type(*, is_link: bool | None) -> str:
        """Return ``"LINK_FLAIR"`` or ``"USER_FLAIR"`` depending on ``is_link`` value."""
        return "LINK_FLAIR" if is_link else "USER_FLAIR"

    def __init__(self, subreddit: models.Subreddit) -> None:
        """Initialize a :class:`.SubredditFlairTemplates` instance.

        :param subreddit: The subreddit whose flair templates to work with.

        .. note::

            This class should not be initialized directly. Instead, obtain an instance
            via: ``reddit.subreddit("test").flair.templates`` or
            ``reddit.subreddit("test").flair.link_templates``.

        """
        self.subreddit = subreddit

    def __iter__(self) -> Iterator[dict[str, str | int | bool | list[dict[str, str]]]]:
        """Abstract method to return flair templates."""
        raise NotImplementedError

    def _add(
        self,
        *,
        allowable_content: str | None = None,
        background_color: str | None = None,
        css_class: str = "",
        is_link: bool | None = None,
        max_emojis: int | None = None,
        mod_only: bool | None = None,
        text: str,
        text_color: str | None = None,
        text_editable: bool = False,
    ) -> None:
        url = API_PATH["flairtemplate_v2"].format(subreddit=self.subreddit)
        data = {
            "allowable_content": allowable_content,
            "background_color": background_color,
            "css_class": css_class,
            "flair_type": self.flair_type(is_link=is_link),
            "max_emojis": max_emojis,
            "mod_only": bool(mod_only),
            "text": text,
            "text_color": text_color,
            "text_editable": bool(text_editable),
        }
        self.subreddit._reddit.post(url, data=data)

    def _clear(self, *, is_link: bool | None = None) -> None:
        url = API_PATH["flairtemplateclear"].format(subreddit=self.subreddit)
        self.subreddit._reddit.post(url, data={"flair_type": self.flair_type(is_link=is_link)})

    def _reorder(self, flair_list: list, *, is_link: bool | None = None) -> None:
        url = API_PATH["flairtemplatereorder"].format(subreddit=self.subreddit)
        self.subreddit._reddit.patch(
            url,
            json=flair_list,
            params={
                "flair_type": self.flair_type(is_link=is_link),
                "subreddit": self.subreddit.display_name,
            },
        )

    def delete(self, template_id: str) -> None:
        """Remove a flair template provided by ``template_id``.

        For example, to delete the first :class:`.Redditor` flair template listed, try:

        .. code-block:: python

            template_info = list(subreddit.flair.templates)[0]
            subreddit.flair.templates.delete(template_info["id"])

        """
        url = API_PATH["flairtemplatedelete"].format(subreddit=self.subreddit)
        self.subreddit._reddit.post(url, data={"flair_template_id": template_id})

    def update(
        self,
        template_id: str,
        *,
        allowable_content: str | None = None,
        background_color: str | None = None,
        css_class: str | None = None,
        fetch: bool = True,
        max_emojis: int | None = None,
        mod_only: bool | None = None,
        text: str | None = None,
        text_color: str | None = None,
        text_editable: bool | None = None,
    ) -> None:
        """Update the flair template provided by ``template_id``.

        :param template_id: The flair template to update. If not valid then an exception
            will be thrown.
        :param allowable_content: If specified, most be one of ``"all"``, ``"emoji"``,
            or ``"text"`` to restrict content to that type. If set to ``"emoji"`` then
            the ``"text"`` param must be a valid emoji string, for example,
            ``":snoo:"``.
        :param background_color: The flair template's new background color, as a hex
            color.
        :param css_class: The flair template's new css_class (default: ``""``).
        :param fetch: Whether PRAW will fetch existing information on the existing flair
            before updating (default: ``True``).
        :param max_emojis: Maximum emojis in the flair (Reddit defaults this value to
            ``10``).
        :param mod_only: Indicate if the flair can only be used by moderators.
        :param text: The flair template's new text.
        :param text_color: The flair template's new text color, either ``"light"`` or
            ``"dark"``.
        :param text_editable: Indicate if the flair text can be modified for each
            redditor that sets it (default: ``False``).

        .. warning::

            If parameter ``fetch`` is set to ``False``, all parameters not provided will
            be reset to their default (``None`` or ``False``) values.

        For example, to make a user flair template text editable, try:

        .. code-block:: python

            template_info = list(subreddit.flair.templates)[0]
            subreddit.flair.templates.update(
                template_info["id"],
                text=template_info["flair_text"],
                text_editable=True,
            )

        """
        url = API_PATH["flairtemplate_v2"].format(subreddit=self.subreddit)
        data = {
            "allowable_content": allowable_content,
            "background_color": background_color,
            "css_class": css_class,
            "flair_template_id": template_id,
            "max_emojis": max_emojis,
            "mod_only": mod_only,
            "text": text,
            "text_color": text_color,
            "text_editable": text_editable,
        }
        if fetch:
            existing_data_ = [template for template in iter(self) if template["id"] == template_id]
            if len(existing_data_) != 1:
                raise InvalidFlairTemplateID(template_id)
            existing_data = existing_data_[0]
            for key, value in existing_data.items():
                if data.get(key) is None:
                    data[key] = value
        self.subreddit._reddit.post(url, data=data)


class SubredditLinkFlairTemplates(SubredditFlairTemplates):
    """Provide functions to interact with link flair templates."""

    def __iter__(
        self,
    ) -> Iterator[dict[str, str | int | bool | list[dict[str, str]]]]:
        """Iterate through the link flair templates as a moderator.

        For example:

        .. code-block:: python

            for template in reddit.subreddit("test").flair.link_templates:
                print(template)

        """
        url = API_PATH["link_flair"].format(subreddit=self.subreddit)
        yield from self.subreddit._reddit.get(url)

    def add(
        self,
        text: str,
        *,
        allowable_content: str | None = None,
        background_color: str | None = None,
        css_class: str = "",
        max_emojis: int | None = None,
        mod_only: bool | None = None,
        text_color: str | None = None,
        text_editable: bool = False,
    ) -> None:
        """Add a link flair template to the associated subreddit.

        :param text: The flair template's text.
        :param allowable_content: If specified, most be one of ``"all"``, ``"emoji"``,
            or ``"text"`` to restrict content to that type. If set to ``"emoji"`` then
            the ``"text"`` param must be a valid emoji string, for example,
            ``":snoo:"``.
        :param background_color: The flair template's new background color, as a hex
            color.
        :param css_class: The flair template's css_class (default: ``""``).
        :param max_emojis: Maximum emojis in the flair (Reddit defaults this value to
            ``10``).
        :param mod_only: Indicate if the flair can only be used by moderators.
        :param text_color: The flair template's new text color, either ``"light"`` or
            ``"dark"``.
        :param text_editable: Indicate if the flair text can be modified for each
            redditor that sets it (default: ``False``).

        For example, to add an editable link flair try:

        .. code-block:: python

            reddit.subreddit("test").flair.link_templates.add(
                "PRAW",
                css_class="praw",
                text_editable=True,
            )

        """
        self._add(
            allowable_content=allowable_content,
            background_color=background_color,
            css_class=css_class,
            is_link=True,
            max_emojis=max_emojis,
            mod_only=mod_only,
            text=text,
            text_color=text_color,
            text_editable=text_editable,
        )

    def clear(self) -> None:
        """Remove all link flair templates from the subreddit.

        For example:

        .. code-block:: python

            reddit.subreddit("test").flair.link_templates.clear()

        """
        self._clear(is_link=True)

    def reorder(self, flair_list: list[str]) -> None:
        """Reorder a list of flairs.

        :param flair_list: A list of flair IDs.

        For example, to reverse the order of the link flair list try:

        .. code-block:: python

            subreddit = reddit.subreddit("test")
            flairs = [flair["id"] for flair in subreddit.flair.link_templates]
            subreddit.flair.link_templates.reorder(list(reversed(flairs)))

        """
        self._reorder(flair_list, is_link=True)

    def user_selectable(
        self,
    ) -> Iterator[dict[str, str | bool]]:
        """Iterate through the link flair templates as a regular user.

        For example:

        .. code-block:: python

            for template in reddit.subreddit("test").flair.link_templates.user_selectable():
                print(template)

        """
        url = API_PATH["flairselector"].format(subreddit=self.subreddit)
        yield from self.subreddit._reddit.post(url, data={"is_newlink": True})["choices"]


class SubredditRedditorFlairTemplates(SubredditFlairTemplates):
    """Provide functions to interact with :class:`.Redditor` flair templates."""

    def __iter__(
        self,
    ) -> Iterator[dict[str, str | int | bool | list[dict[str, str]]]]:
        """Iterate through the user flair templates.

        For example:

        .. code-block:: python

            for template in reddit.subreddit("test").flair.templates:
                print(template)

        """
        url = API_PATH["user_flair"].format(subreddit=self.subreddit)
        params: dict[str, str | int] = {"unique": self.subreddit._reddit._next_unique}
        yield from self.subreddit._reddit.get(url, params=params)

    def add(
        self,
        text: str,
        *,
        allowable_content: str | None = None,
        background_color: str | None = None,
        css_class: str = "",
        max_emojis: int | None = None,
        mod_only: bool | None = None,
        text_color: str | None = None,
        text_editable: bool = False,
    ) -> None:
        """Add a redditor flair template to the associated subreddit.

        :param text: The flair template's text.
        :param allowable_content: If specified, most be one of ``"all"``, ``"emoji"``,
            or ``"text"`` to restrict content to that type. If set to ``"emoji"`` then
            the ``"text"`` param must be a valid emoji string, for example,
            ``":snoo:"``.
        :param background_color: The flair template's new background color, as a hex
            color.
        :param css_class: The flair template's css_class (default: ``""``).
        :param max_emojis: Maximum emojis in the flair (Reddit defaults this value to
            ``10``).
        :param mod_only: Indicate if the flair can only be used by moderators.
        :param text_color: The flair template's new text color, either ``"light"`` or
            ``"dark"``.
        :param text_editable: Indicate if the flair text can be modified for each
            redditor that sets it (default: ``False``).

        For example, to add an editable redditor flair try:

        .. code-block:: python

            reddit.subreddit("test").flair.templates.add(
                "PRAW",
                css_class="praw",
                text_editable=True,
            )

        """
        self._add(
            allowable_content=allowable_content,
            background_color=background_color,
            css_class=css_class,
            is_link=False,
            max_emojis=max_emojis,
            mod_only=mod_only,
            text=text,
            text_color=text_color,
            text_editable=text_editable,
        )

    def clear(self) -> None:
        """Remove all :class:`.Redditor` flair templates from the subreddit.

        For example:

        .. code-block:: python

            reddit.subreddit("test").flair.templates.clear()

        """
        self._clear(is_link=False)

    def reorder(self, flair_list: list[str]) -> None:
        """Reorder a list of flairs.

        :param flair_list: A list of flair IDs.

        For example, to reverse the order of the :class:`.Redditor` flair templates list
        try:

        .. code-block:: python

            subreddit = reddit.subreddit("test")
            flairs = [flair["id"] for flair in subreddit.flair.templates]
            subreddit.flair.templates.reorder(list(reversed(flairs)))

        """
        self._reorder(flair_list, is_link=False)



"""Provide the BaseList class."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar

from praw.models.base import PRAWBase

if TYPE_CHECKING:
    from collections.abc import Iterator

    import praw


class BaseList(PRAWBase):
    """An abstract class to coerce a list into a :class:`.PRAWBase`."""

    CHILD_ATTRIBUTE: ClassVar[str | None] = None

    def __contains__(self, item: Any) -> bool:
        """Test if item exists in the list."""
        return item in getattr(self, self._child_attribute())

    def __getitem__(self, index: int) -> Any:
        """Return the item at position index in the list."""
        return getattr(self, self._child_attribute())[index]

    def __init__(self, reddit: praw.Reddit, _data: dict[str, Any]) -> None:
        """Initialize a :class:`.BaseList` instance.

        :param reddit: An instance of :class:`.Reddit`.

        """
        super().__init__(reddit, _data=_data)

        child_list = getattr(self, self._child_attribute())
        for index, item in enumerate(child_list):
            child_list[index] = reddit._objector.objectify(data=item)

    def __iter__(self) -> Iterator[Any]:
        """Return an iterator to the list."""
        return getattr(self, self._child_attribute()).__iter__()

    def __len__(self) -> int:
        """Return the number of items in the list."""
        return len(getattr(self, self._child_attribute()))

    def __str__(self) -> str:
        """Return a string representation of the list."""
        return str(getattr(self, self._child_attribute()))

    def _child_attribute(self) -> str:
        """Return ``CHILD_ATTRIBUTE``, ensuring it has been set by a subclass."""
        if self.CHILD_ATTRIBUTE is None:
            msg = "BaseList must be extended."
            raise NotImplementedError(msg)
        return self.CHILD_ATTRIBUTE



"""Provides the User class."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from prawcore import Conflict

from praw.const import API_PATH
from praw.exceptions import ReadOnlyException
from praw.models import Preferences
from praw.models.base import PRAWBase
from praw.models.listing.generator import ListingGenerator
from praw.models.reddit.redditor import Redditor
from praw.models.reddit.subreddit import Subreddit
from praw.util.cache import cachedproperty

if TYPE_CHECKING:
    from collections.abc import Iterator

    import praw
    from praw import models


class User(PRAWBase):
    """The :class:`.User` class provides methods for the currently authenticated user."""

    @cachedproperty
    def preferences(self) -> models.Preferences:
        """Get an instance of :class:`.Preferences`.

        The preferences can be accessed as a ``dict`` like so:

        .. code-block:: python

            preferences = reddit.user.preferences()
            print(preferences["show_link_flair"])

        Preferences can be updated via:

        .. code-block:: python

            reddit.user.preferences.update(show_link_flair=True)

        The :meth:`.Preferences.update` method returns the new state of the preferences
        as a ``dict``, which can be used to check whether a change went through. Changes
        with invalid types or parameter names fail silently.

        .. code-block:: python

            original_preferences = reddit.user.preferences()
            new_preferences = reddit.user.preferences.update(invalid_param=123)
            print(original_preferences == new_preferences)  # True, no change

        """
        return Preferences(self._reddit)

    def __init__(self, reddit: praw.Reddit) -> None:
        """Initialize an :class:`.User` instance.

        This class is intended to be interfaced with through ``reddit.user``.

        """
        super().__init__(reddit, _data=None)

    def blocked(self) -> list[models.Redditor]:
        r"""Return a :class:`.RedditorList` of blocked :class:`.Redditor`\ s."""
        return self._reddit.get(API_PATH["blocked"])

    def contributor_subreddits(self, **generator_kwargs: Any) -> Iterator[models.Subreddit]:
        r"""Return a :class:`.ListingGenerator` of contributor :class:`.Subreddit`\ s.

        These are subreddits in which the user is an approved user.

        Additional keyword arguments are passed in the initialization of
        :class:`.ListingGenerator`.

        To print a list of the subreddits that you are an approved user in, try:

        .. code-block:: python

            for subreddit in reddit.user.contributor_subreddits(limit=None):
                print(str(subreddit))

        """
        return ListingGenerator(self._reddit, API_PATH["my_contributor"], **generator_kwargs)

    def friends(self, *, user: str | models.Redditor | None = None) -> list[models.Redditor] | models.Redditor:
        r"""Return a :class:`.RedditorList` of friends or a :class:`.Redditor` in the friends list.

        :param user: Checks to see if you are friends with the redditor. Either an
            instance of :class:`.Redditor` or a string can be given.

        :returns: A list of :class:`.Redditor`\ s, or a single :class:`.Redditor` if
            ``user`` is specified. The :class:`.Redditor` instance(s) returned also has
            friend attributes.

        :raises: An instance of :class:`.RedditAPIException` if you are not friends with
            the specified :class:`.Redditor`.

        """
        endpoint = API_PATH["friends"] if user is None else API_PATH["friend_v1"].format(user=str(user))
        return self._reddit.get(endpoint)

    def karma(self) -> dict[models.Subreddit, dict[str, int]]:
        r"""Return a dictionary mapping :class:`.Subreddit`\ s to their karma.

        The returned dict contains subreddits as keys. Each subreddit key contains a
        sub-dict that have keys for ``comment_karma`` and ``link_karma``. The dict is
        sorted in descending karma order.

        .. note::

            Each key of the main dict is an instance of :class:`.Subreddit`. It is
            recommended to iterate over the dict in order to retrieve the values,
            preferably through :py:meth:`dict.items`.

        """
        karma_map = {}
        for row in self._reddit.get(API_PATH["karma"])["data"]:
            subreddit = Subreddit(self._reddit, row["sr"])
            del row["sr"]
            karma_map[subreddit] = row
        return karma_map

    def me(self, *, use_cache: bool = True) -> models.Redditor | None:
        """Return a :class:`.Redditor` instance for the authenticated user.

        :param use_cache: When ``True``, and if this function has been previously
            called, returned the cached version (default: ``True``).

        .. note::

            If you change the :class:`.Reddit` instance's authorization, you might want
            to refresh the cached value. Prefer using separate :class:`.Reddit`
            instances, however, for distinct authorizations.

        """
        if self._reddit.read_only:
            msg = "`user.me()` does not work in read_only mode"
            raise ReadOnlyException(msg)
        if "_me" not in self.__dict__ or not use_cache:
            user_data = self._reddit.get(API_PATH["me"])
            self._me = Redditor(self._reddit, _data=user_data)
        return self._me

    def moderator_subreddits(self, **generator_kwargs: Any) -> Iterator[models.Subreddit]:
        """Return a :class:`.ListingGenerator` subreddits that the user moderates.

        Additional keyword arguments are passed in the initialization of
        :class:`.ListingGenerator`.

        To print a list of the names of the subreddits you moderate, try:

        .. code-block:: python

            for subreddit in reddit.user.moderator_subreddits(limit=None):
                print(str(subreddit))

        .. seealso::

            :meth:`.Redditor.moderated`

        """
        return ListingGenerator(self._reddit, API_PATH["my_moderator"], **generator_kwargs)

    def multireddits(self) -> list[models.Multireddit]:
        r"""Return a list of :class:`.Multireddit`\ s belonging to the user."""
        return self._reddit.get(API_PATH["my_multireddits"])

    def pin(
        self, submission: models.Submission, *, num: int | None = None, state: bool = True
    ) -> models.Submission | None:
        """Set the pin state of a submission on the authenticated user's profile.

        :param submission: An instance of :class:`.Submission` that will be
            pinned/unpinned.
        :param num: If specified, the slot in which the submission will be pinned into.
            If there is a submission already in the specified slot, it will be replaced.
            If ``None`` or there is not a submission in the specified slot, the first
            available slot will be used (default: ``None``). If all slots are used the
            following will occur:

            - Old Reddit:

              1. The submission in the last slot will be unpinned.
              2. The remaining pinned submissions will be shifted down a slot.
              3. The new submission will be pinned in the first slot.

            - New Reddit:

              1. The submission in the first slot will be unpinned.
              2. The remaining pinned submissions will be shifted up a slot.
              3. The new submission will be pinned in the last slot.

            .. note::

                At the time of writing (10/22/2021), there are 4 pin slots available and
                pins are in reverse order on old Reddit. If ``num`` is an invalid value,
                Reddit will ignore it and the same behavior will occur as if ``num`` is
                ``None``.

        :param state: ``True`` pins the submission, ``False`` unpins (default:
            ``True``).

        :returns: The pinned submission.

        :raises: ``prawcore.BadRequest`` when pinning a removed or deleted submission.
        :raises: ``prawcore.Forbidden`` when pinning a submission the authenticated user
            is not the author of.

        .. code-block:: python

            submission = next(reddit.user.me().submissions.new())
            reddit.user.pin(submission)

        """
        data = {
            "id": submission.fullname,
            "num": num,
            "state": state,
            "to_profile": True,
        }
        try:
            return self._reddit.post(API_PATH["sticky_submission"], data=data)
        except Conflict:
            pass

    def subreddits(self, **generator_kwargs: Any) -> Iterator[models.Subreddit]:
        r"""Return a :class:`.ListingGenerator` of :class:`.Subreddit`\ s the user is subscribed to.

        Additional keyword arguments are passed in the initialization of
        :class:`.ListingGenerator`.

        To print a list of the subreddits that you are subscribed to, try:

        .. code-block:: python

            for subreddit in reddit.user.subreddits(limit=None):
                print(str(subreddit))

        """
        return ListingGenerator(self._reddit, API_PATH["my_subreddits"], **generator_kwargs)

    def trusted(self) -> list[models.Redditor]:
        r"""Return a :class:`.RedditorList` of trusted :class:`.Redditor`\ s.

        To display the usernames of your trusted users and the times at which you
        decided to trust them, try:

        .. code-block:: python

            trusted_users = reddit.user.trusted()
            for user in trusted_users:
                print(f"User: {user.name}, time: {user.date}")

        """
        return self._reddit.get(API_PATH["trusted"])



"""Provide the :class:`.Subreddit` class and its helper classes."""

from __future__ import annotations

from praw.models.reddit.subreddit.filters import SubredditFilters
from praw.models.reddit.subreddit.flair import (
    SubredditFlair,
    SubredditFlairTemplates,
    SubredditLinkFlairTemplates,
    SubredditRedditorFlairTemplates,
)
from praw.models.reddit.subreddit.moderation import (
    SubredditModeration,
    SubredditModerationStream,
)
from praw.models.reddit.subreddit.modmail import Modmail
from praw.models.reddit.subreddit.quarantine import SubredditQuarantine
from praw.models.reddit.subreddit.relationship import (
    ContributorRelationship,
    ModeratorRelationship,
    SubredditRelationship,
)
from praw.models.reddit.subreddit.stream import SubredditStream
from praw.models.reddit.subreddit.stylesheet import SubredditStylesheet
from praw.models.reddit.subreddit.subreddit import Subreddit
from praw.models.reddit.subreddit.wiki import SubredditWiki

__all__ = [
    "ContributorRelationship",
    "ModeratorRelationship",
    "Modmail",
    "Subreddit",
    "SubredditFilters",
    "SubredditFlair",
    "SubredditFlairTemplates",
    "SubredditLinkFlairTemplates",
    "SubredditModeration",
    "SubredditModerationStream",
    "SubredditQuarantine",
    "SubredditRedditorFlairTemplates",
    "SubredditRelationship",
    "SubredditStream",
    "SubredditStylesheet",
    "SubredditWiki",
]



"""Provide the Rule class."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, SupportsIndex
from urllib.parse import quote

from praw.const import API_PATH
from praw.exceptions import ClientException
from praw.models.reddit.base import RedditBase
from praw.models.reddit.mixins import CreatedMixin
from praw.util import cachedproperty

if TYPE_CHECKING:
    from collections.abc import Iterator

    import praw
    from praw import models


class Rule(CreatedMixin, RedditBase):
    """An individual :class:`.Rule` object.

    .. include:: ../../typical_attributes.rst

    ==================== =============================================================
    Attribute            Description
    ==================== =============================================================
    ``created_utc``      Time the rule was created, represented in `Unix Time`_.
    ``description``      The description of the rule, if provided, otherwise a blank
                         string.
    ``kind``             The kind of rule. Can be ``"link"``, ``comment"``, or
                         ``"all"``.
    ``priority``         Represents where the rule is ranked. For example, the first
                         rule is at priority ``0``. Serves as an index number on the
                         list of rules.
    ``short_name``       The name of the rule.
    ``violation_reason`` The reason that is displayed on the report menu for the rule.
    ==================== =============================================================

    .. _unix time: https://en.wikipedia.org/wiki/Unix_time

    """

    STR_FIELD = "short_name"

    @cachedproperty
    def mod(self) -> RuleModeration:
        """Contain methods used to moderate rules.

        To delete ``"No spam"`` from r/test try:

        .. code-block:: python

            reddit.subreddit("test").rules["No spam"].mod.delete()

        To update ``"No spam"`` from r/test try:

        .. code-block:: python

            reddit.subreddit("test").removal_reasons["No spam"].mod.update(
                description="Don't do this!", violation_reason="Spam post"
            )

        """
        return RuleModeration(self)

    def __getattribute__(self, attribute: str) -> Any:
        """Get the value of an attribute."""
        value = super().__getattribute__(attribute)
        if attribute == "subreddit" and value is None:
            msg = "The Rule is missing a subreddit. File a bug report at PRAW."
            raise ValueError(msg)
        return value

    def __init__(
        self,
        reddit: praw.Reddit,
        subreddit: models.Subreddit | None = None,
        short_name: str | None = None,
        _data: dict[str, str] | None = None,
    ) -> None:
        """Initialize a :class:`.Rule` instance."""
        if (short_name, _data).count(None) != 1:
            msg = "Either short_name or _data needs to be given."
            raise ValueError(msg)
        if short_name:
            self.short_name = short_name
        # Note: The subreddit parameter can be None, because the objector does not know
        # this info. In that case, it is the responsibility of the caller to set the
        # `subreddit` property on the returned value.
        self.subreddit = subreddit
        super().__init__(reddit, _data=_data)

    def _fetch(self) -> None:
        assert self.subreddit is not None
        for rule in self.subreddit.rules:
            if rule.short_name == self.short_name:
                self.__dict__.update(rule.__dict__)
                super()._fetch()
                return
        msg = f"Subreddit {self.subreddit} does not have the rule {self.short_name}"
        raise ClientException(msg)


class RuleModeration:
    """Contain methods used to moderate rules.

    To delete ``"No spam"`` from r/test try:

    .. code-block:: python

        reddit.subreddit("test").rules["No spam"].mod.delete()

    To update ``"No spam"`` from r/test try:

    .. code-block:: python

        reddit.subreddit("test").removal_reasons["No spam"].mod.update(
            description="Don't do this!", violation_reason="Spam post"
        )

    """

    def __init__(self, rule: models.Rule) -> None:
        """Initialize a :class:`.RuleModeration` instance."""
        self.rule = rule

    def delete(self) -> None:
        """Delete a rule from this subreddit.

        To delete ``"No spam"`` from r/test try:

        .. code-block:: python

            reddit.subreddit("test").rules["No spam"].mod.delete()

        """
        data = {
            "r": str(self.rule.subreddit),
            "short_name": self.rule.short_name,
        }
        self.rule._reddit.post(API_PATH["remove_subreddit_rule"], data=data)

    def update(
        self,
        *,
        description: str | None = None,
        kind: str | None = None,
        short_name: str | None = None,
        violation_reason: str | None = None,
    ) -> models.Rule:
        """Update the rule from this subreddit.

        .. note::

            Existing values will be used for any unspecified arguments.

        :param description: The new description for the rule. Can be empty.
        :param kind: The kind of item that the rule applies to. One of ``"link"``,
            ``"comment"``, or ``"all"``.
        :param short_name: The name of the rule.
        :param violation_reason: The reason that is shown on the report menu.

        :returns: A Rule object containing the updated values.

        To update ``"No spam"`` from r/test try:

        .. code-block:: python

            reddit.subreddit("test").removal_reasons["No spam"].mod.update(
                description="Don't do this!", violation_reason="Spam post"
            )

        """
        data = {
            "old_short_name": self.rule.short_name,
            "r": str(self.rule.subreddit),
        }
        for name, value in {
            "description": description,
            "kind": kind,
            "short_name": short_name,
            "violation_reason": violation_reason,
        }.items():
            data[name] = getattr(self.rule, name) if value is None else value
        updated_rule = self.rule._reddit.post(API_PATH["update_subreddit_rule"], data=data)[0]
        updated_rule.subreddit = self.rule.subreddit
        return updated_rule


class SubredditRules:
    """Provide a set of functions to access a :class:`.Subreddit`'s rules.

    For example, to list all the rules for a subreddit:

    .. code-block:: python

        for rule in reddit.subreddit("test").rules:
            print(rule)

    Moderators can also add rules to the subreddit. For example, to make a rule called
    ``"No spam"`` in r/test:

    .. code-block:: python

        reddit.subreddit("test").rules.mod.add(
            short_name="No spam", kind="all", description="Do not spam. Spam bad"
        )

    """

    @cachedproperty
    def _rule_list(self) -> list[Rule]:
        """Get a list of :class:`.Rule` objects.

        :returns: A list of instances of :class:`.Rule`.

        """
        rule_list = self._reddit.get(API_PATH["rules"].format(subreddit=self.subreddit))
        for rule in rule_list:
            rule.subreddit = self.subreddit
        return rule_list

    @cachedproperty
    def mod(self) -> SubredditRulesModeration:
        """Contain methods to moderate subreddit rules as a whole.

        To add rule ``"No spam"`` to r/test try:

        .. code-block:: python

            reddit.subreddit("test").rules.mod.add(
                short_name="No spam", kind="all", description="Do not spam. Spam bad"
            )

        To move the fourth rule to the first position, and then to move the prior first
        rule to where the third rule originally was in r/test:

        .. code-block:: python

            subreddit = reddit.subreddit("test")
            rules = list(subreddit.rules)
            new_rules = rules[3:4] + rules[1:3] + rules[0:1] + rules[4:]
            # Alternate: [rules[3]] + rules[1:3] + [rules[0]] + rules[4:]
            new_rule_list = subreddit.rules.mod.reorder(new_rules)

        """
        return SubredditRulesModeration(self)

    def __getitem__(self, short_name: SupportsIndex) -> models.Rule:
        """Return the :class:`.Rule` for the subreddit with short_name ``short_name``.

        :param short_name: The short_name of the rule, or the rule number.

        .. note::

            Rules fetched using a specific rule name are lazily loaded, so you might
            have to access an attribute to get all the expected attributes.

        This method is to be used to fetch a specific rule, like so:

        .. code-block:: python

            rule_name = "No spam"
            rule = reddit.subreddit("test").rules[rule_name]
            print(rule)

        You can also fetch a numbered rule of a subreddit.

        Rule numbers start at ``0``, so the first rule is at index ``0``, and the second
        rule is at index ``1``, and so on.

        :raises: :py:class:`IndexError` if a rule of a specific number does not exist.

        .. note::

            You can use negative indexes, such as ``-1``, to get the last rule. You can
            also use slices, to get a subset of rules, such as the last three rules with
            ``rules[-3:]``.

        For example, to fetch the second rule of r/test:

        .. code-block:: python

            rule = reddit.subreddit("test").rules[1]

        To get the last three rules in a subreddit:

        .. code-block:: python

            reasons = reddit.subreddit("test").rules[-3:]
            for rule in rules:
                print(rule)

        """
        if not isinstance(short_name, str):
            return self._rule_list[short_name]
        return Rule(self._reddit, short_name=short_name, subreddit=self.subreddit)

    def __init__(self, subreddit: models.Subreddit) -> None:
        """Initialize a :class:`.SubredditRules` instance.

        :param subreddit: The subreddit whose rules to work with.

        """
        self.subreddit = subreddit
        self._reddit = subreddit._reddit

    def __iter__(self) -> Iterator[models.Rule]:
        """Iterate through the rules of the subreddit.

        :returns: An iterator containing all the rules of a subreddit.

        This method is used to discover all rules for a subreddit.

        For example, to get the rules for r/test:

        .. code-block:: python

            for rule in reddit.subreddit("test").rules:
                print(rule)

        """
        return iter(self._rule_list)


class SubredditRulesModeration:
    """Contain methods to moderate subreddit rules as a whole.

    To add rule ``"No spam"`` to r/test try:

    .. code-block:: python

        reddit.subreddit("test").rules.mod.add(
            short_name="No spam", kind="all", description="Do not spam. Spam bad"
        )

    To move the fourth rule to the first position, and then to move the prior first rule
    to where the third rule originally was in r/test:

    .. code-block:: python

        subreddit = reddit.subreddit("test")
        rules = list(subreddit.rules)
        new_rules = rules[3:4] + rules[1:3] + rules[0:1] + rules[4:]
        # Alternate: [rules[3]] + rules[1:3] + [rules[0]] + rules[4:]
        new_rule_list = subreddit.rules.mod.reorder(new_rules)

    """

    def __init__(self, subreddit_rules: SubredditRules) -> None:
        """Initialize a :class:`.SubredditRulesModeration` instance."""
        self.subreddit_rules = subreddit_rules

    def add(
        self,
        *,
        description: str = "",
        kind: str,
        short_name: str,
        violation_reason: str | None = None,
    ) -> models.Rule:
        """Add a removal reason to this subreddit.

        :param description: The description for the rule.
        :param kind: The kind of item that the rule applies to. One of ``"link"``,
            ``"comment"``, or ``"all"``.
        :param short_name: The name of the rule.
        :param violation_reason: The reason that is shown on the report menu. If a
            violation reason is not specified, the short name will be used as the
            violation reason.

        :returns: The added :class:`.Rule`.

        To add rule ``"No spam"`` to r/test try:

        .. code-block:: python

            reddit.subreddit("test").rules.mod.add(
                short_name="No spam", kind="all", description="Do not spam. Spam bad"
            )

        """
        data = {
            "description": description,
            "kind": kind,
            "r": str(self.subreddit_rules.subreddit),
            "short_name": short_name,
            "violation_reason": (short_name if violation_reason is None else violation_reason),
        }
        new_rule = self.subreddit_rules._reddit.post(API_PATH["add_subreddit_rule"], data=data)[0]
        new_rule.subreddit = self.subreddit_rules.subreddit
        return new_rule

    def reorder(self, rule_list: list[models.Rule]) -> list[models.Rule]:
        """Reorder the rules of a subreddit.

        :param rule_list: The list of rules, in the wanted order. Each index of the list
            indicates the position of the rule.

        :returns: A list containing the rules in the specified order.

        For example, to move the fourth rule to the first position, and then to move the
        prior first rule to where the third rule originally was in r/test:

        .. code-block:: python

            subreddit = reddit.subreddit("test")
            rules = list(subreddit.rules)
            new_rules = rules[3:4] + rules[1:3] + rules[0:1] + rules[4:]
            # Alternate: [rules[3]] + rules[1:3] + [rules[0]] + rules[4:]
            new_rule_list = subreddit.rules.mod.reorder(new_rules)

        """
        order_string = quote(",".join([rule.short_name for rule in rule_list]), safe=",")
        data = {
            "new_rule_order": order_string,
            "r": str(self.subreddit_rules.subreddit),
        }
        response = self.subreddit_rules._reddit.post(API_PATH["reorder_subreddit_rules"], data=data)
        for rule in response:
            rule.subreddit = self.subreddit_rules.subreddit
        return response
